"""Draw hypergraphs and simplicial complexes with matplotlib."""

from collections.abc import Iterable
from inspect import signature
from itertools import combinations

import matplotlib.pyplot as plt
import numpy as np
from matplotlib import cm
from matplotlib.colors import LinearSegmentedColormap, ListedColormap

from .. import convert
from ..classes import Hypergraph, SimplicialComplex, max_edge_order
from ..exception import XGIError
from ..stats import EdgeStat, NodeStat
from .layout import barycenter_spring_layout

__all__ = [
    "draw",
    "draw_xgi_nodes",
    "draw_xgi_hyperedges",
    "draw_xgi_simplices",
    "draw_node_labels",
    "draw_hyperedge_labels",
]


def draw(
    H,
    pos=None,
    ax=None,
    dyad_color="black",
    dyad_lw=1.5,
    edge_fc=None,
    node_fc="white",
    node_ec="black",
    node_lw=1,
    node_size=15,
    max_order=None,
    node_labels=False,
    hyperedge_labels=False,
    **kwargs,
):
    """Draw hypergraph or simplicial complex.

    Parameters
    ----
    H : Hypergraph or SimplicialComplex.
    pos : dict (default=None)
        If passed, this dictionary of positions node_id:(x,y) is used for placing the 0-simplices.
        If None (default), use the `barycenter_spring_layout` to compute the positions.
    ax : matplotlib.pyplot.axes (default=None)
    dyad_color : str, dict, iterable, or EdgeStat (default='black')
        Color of the dyadic links.  If str, use the same color for all edges. If a dict, must
        contain (edge_id: color_str) pairs.  If iterable, assume the colors are
        specified in the same order as the edges are found in H.edges. If EdgeStat, use a colormap
        (specified with dyad_color_cmap) associated to it.
    dyad_lw : int, float, dict, iterable, or EdgeStat (default=1.5)
        Line width of edges of order 1 (dyadic links).  If int or float, use the same width for all edges.
        If a dict, must contain (edge_id: width) pairs.  If iterable, assume the widths are
        specified in the same order as the edges are found in H.edges. If EdgeStat, use a monotonic
        linear interpolation defined between min_dyad_lw and max_dyad_lw.
    edge_fc : str, dict, iterable, or EdgeStat (default=None)
        Color of the hyperedges.  If str, use the same color for all nodes.  If a dict, must
        contain (edge_id: color_str) pairs.  If other iterable, assume the colors are
        specified in the same order as the hyperedges are found in H.edges. If EdgeStat,
        use the colormap specified with edge_fc_cmap. If None (default),
        use the H.edges.size.
    node_fc : str, dict, iterable, or NodeStat (default='white')
        Color of the nodes.  If str, use the same color for all nodes.  If a dict, must
        contain (node_id: color_str) pairs.  If other iterable, assume the colors are
        specified in the same order as the nodes are found in H.nodes. If NodeStat,
        use the colormap specified with node_fc_cmap.
    node_ec : str, dict, iterable, or NodeStat (default='black')
        Color of node borders.  If str, use the same color for all nodes.  If a dict, must
        contain (node_id: color_str) pairs.  If other iterable, assume the colors are
        specified in the same order as the nodes are found in H.nodes. If NodeStat,
        use the colormap specified with node_ec_cmap.
    node_lw : int, float, dict, iterable, or NodeStat (default=1)
        Line width of the node borders in pixels.  If int or float, use the same width for all node borders.
        If a dict, must contain (node_id: width) pairs.  If iterable, assume the widths are
        specified in the same order as the nodes are found in H.nodes. If NodeStat, use a monotonic
        linear interpolation defined between min_node_lw and max_node_lw.
    node_size : int, float, dict, iterable, or NodeStat (default=10)
        Radius of the nodes in pixels.  If int or float, use the same radius for all nodes.
        If a dict, must contain (node_id: radius) pairs.  If iterable, assume the radiuses are
        specified in the same order as the nodes are found in H.nodes. If NodeStat, use a monotonic
        linear interpolation defined between min_node_size and max_node_size.
    max_order : int (default=None)
        Maximum of hyperedges to plot. If None (default), plots all orders.
    node_labels : bool, or dict (default=False)
        If True, draw ids on the nodes. If a dict, must contain (node_id: label) pairs.
    hyperedge_labels : bool, or dict (default=False)
        If True, draw ids on the hyperedges. If a dict, must contain (edge_id: label) pairs.
    **kwargs : optional args
        Alternate default values. Values that can be overwritten are the following:
        * min_node_size
        * max_node_size
        * min_node_lw
        * max_node_lw
        * min_dyad_lw
        * max_dyad_lw
        * node_fc_cmap
        * node_ec_cmap
        * dyad_color_cmap
        * edge_fc_cmap

    Examples
    --------
    >>> import xgi
    >>> H = xgi.Hypergraph()
    >>> H.add_edges_from([[1,2,3],[3,4],[4,5,6,7],[7,8,9,10,11]])
    >>> xgi.draw(H, pos=xgi.barycenter_spring_layout(H))

    See Also
    --------
    draw_xgi_nodes
    draw_xgi_hyperedges
    draw_xgi_simplices
    draw_node_labels
    draw_hyperedge_labels
    """
    settings = {
        "min_node_size": 10.0,
        "max_node_size": 30.0,
        "min_dyad_lw": 2.0,
        "max_dyad_lw": 10.0,
        "min_node_lw": 1.0,
        "max_node_lw": 5.0,
        "node_fc_cmap": cm.Reds,
        "node_ec_cmap": cm.Greys,
        "edge_fc_cmap": cm.Blues,
        "dyad_color_cmap": cm.Greys,
    }

    settings.update(kwargs)

    if edge_fc is None:
        edge_fc = H.edges.size

    if pos is None:
        pos = barycenter_spring_layout(H)

    if ax is None:
        ax = plt.gca()
    ax.set_xlim([-1.1, 1.1])
    ax.set_ylim([-1.1, 1.1])
    ax.get_xaxis().set_ticks([])
    ax.get_yaxis().set_ticks([])
    ax.axis("off")

    if not max_order:
        max_order = max_edge_order(H)

    if isinstance(H, SimplicialComplex):
        draw_xgi_simplices(
            H,
            pos,
            ax,
            dyad_color,
            dyad_lw,
            edge_fc,
            max_order,
            settings,
            hyperedge_labels,
            **kwargs,
        )
    elif isinstance(H, Hypergraph):
        draw_xgi_hyperedges(
            H,
            pos,
            ax,
            dyad_color,
            dyad_lw,
            edge_fc,
            max_order,
            settings,
            hyperedge_labels,
            **kwargs,
        )
    else:
        raise XGIError("The input must be a SimplicialComplex or Hypergraph")

    draw_xgi_nodes(
        H,
        pos,
        ax,
        node_fc,
        node_ec,
        node_lw,
        node_size,
        max_order,
        settings,
        node_labels,
        **kwargs,
    )


def draw_xgi_nodes(
    H,
    pos,
    ax,
    node_fc,
    node_ec,
    node_lw,
    node_size,
    zorder,
    settings,
    node_labels,
    **kwargs,
):
    """Draw the nodes of a hypergraph

    Parameters
    ----------
    H : Hypergraph or SimplicialComplex
        Higher-order network to plot.
    ax : matplotlib.pyplot.axes
    pos : dict
        Dictionary of positions node_id:(x,y) of every node.
    node_fc : str, dict, iterable, or NodeStat
        Color of the nodes.  If str, use the same color for all nodes.  If a dict, must
        contain (node_id: color_str) pairs.  If other iterable, assume the colors are
        specified in the same order as the nodes are found in H.nodes. If NodeStat,
        use the colormap specified with node_fc_cmap.
    node_ec : str, dict, iterable, or NodeStat (default='black')
        Color of node borders.  If str, use the same color for all nodes.  If a dict, must
        contain (node_id: color_str) pairs.  If other iterable, assume the colors are
        specified in the same order as the nodes are found in H.nodes. If NodeStat,
        use the colormap specified with node_ec_cmap.
    node_lw : int, float, dict, iterable, or EdgeStat (default=1)
        Line width of the node borders in pixels.  If int or float, use the same width for all node borders.
        If a dict, must contain (node_id: width) pairs.  If iterable, assume the widths are
        specified in the same order as the nodes are found in H.nodes. If NodeStat, use a monotonic
        linear interpolation defined between min_node_lw and max_node_lw.
    node_size : int, float, dict, iterable, or NodeStat (default=10)
        Radius of the nodes in pixels.  If int or float, use the same radius for all nodes.
        If a dict, must contain (node_id: radius) pairs.  If iterable, assume the radiuses are
        specified in the same order as the nodes are found in H.nodes. If NodeStat, use a monotonic
        linear interpolation defined between min_node_size and max_node_size.
    zorder : int
        The layer on which to draw the nodes.
    node_labels : bool, or dict
        If True, draw ids on the nodes. If a dict, must contain (node_id: label) pairs.
    settings : dict
        Default parameters. Keys that may be useful to override default settings:
        * min_node_size
        * max_node_size
        * min_node_lw
        * max_node_lw
        * node_fc_cmap
        * node_ec_cmap
    kwargs : optional keywords
        See `draw_node_labels` for a description of optional keywords.

    See Also
    --------
    draw
    draw_xgi_hyperedges
    draw_xgi_simplices
    draw_node_labels
    draw_hyperedge_labels
    """
    # Note Iterable covers lists, tuples, ranges, generators, np.ndarrays, etc
    node_fc = _color_arg_to_dict(node_fc, H.nodes, settings["node_fc_cmap"])
    node_ec = _color_arg_to_dict(node_ec, H.nodes, settings["node_ec_cmap"])
    node_lw = _scalar_arg_to_dict(
        node_lw,
        H.nodes,
        settings["min_node_lw"],
        settings["max_node_lw"],
    )
    node_size = _scalar_arg_to_dict(
        node_size, H.nodes, settings["min_node_size"], settings["max_node_size"]
    )

    x, y, s, c, ec, lw, = zip(
        *[
            (
                pos[i][0],
                pos[i][1],
                node_size[i] ** 2,
                node_fc[i],
                node_ec[i],
                node_lw[i],
            )
            for i in H.nodes
        ]
    )
    ax.scatter(x=x, y=y, s=s, c=c, edgecolors=ec, linewidths=lw, zorder=zorder)

    if node_labels:
        # Get all valid keywords by inspecting the signatures of draw_node_labels
        valid_label_kwds = signature(draw_node_labels).parameters.keys()
        # Remove the arguments of this function (draw_networkx)
        valid_label_kwds = valid_label_kwds - {"H", "pos", "ax", "node_labels"}
        if any([k not in valid_label_kwds for k in kwargs]):
            invalid_args = ", ".join([k for k in kwargs if k not in valid_label_kwds])
            raise ValueError(f"Received invalid argument(s): {invalid_args}")
        label_kwds = {k: v for k, v in kwargs.items() if k in valid_label_kwds}
        draw_node_labels(H, pos, node_labels, ax_nodes=ax, **label_kwds)


def draw_xgi_hyperedges(
    H,
    pos,
    ax,
    dyad_color,
    dyad_lw,
    edge_fc,
    max_order,
    settings,
    hyperedge_labels,
    **kwargs,
):
    """Draw hyperedges.

    Parameters
    ----------
    H : Hypergraph
    ax : matplotlib.pyplot.axes
    pos : dict
        Dictionary of positions node_id:(x,y) of every node.
    dyad_color : str, dict, iterable, or EdgeStat
        Color of the dyadic links.  If str, use the same color for all edges. If a dict, must
        contain (edge_id: color_str) pairs.  If iterable, assume the colors are
        specified in the same order as the edges are found in H.edges. If EdgeStat, use a colormap
        (specified with dyad_color_cmap) associated to it.
    dyad_lw : int, float, dict, iterable, or EdgeStat
        Line width of edges of order 1 (dyadic links).  If int or float, use the same width for all edges.
        If a dict, must contain (edge_id: width) pairs.  If iterable, assume the widths are
        specified in the same order as the edges are found in H.edges. If EdgeStat, use a monotonic
        linear interpolation defined between min_dyad_lw and max_dyad_lw.
    edge_fc : str, dict, iterable, or EdgeStat (default=None)
        Color of the hyperedges.  If str, use the same color for all nodes.  If a dict, must
        contain (edge_id: color_str) pairs.  If other iterable, assume the colors are
        specified in the same order as the hyperedges are found in H.edges. If EdgeStat,
        use the colormap specified with edge_fc_cmap.
    max_order : int
        Maximum of hyperedges to plot.
    hyperedge_labels : bool, or dict
        If True, draw ids on the hyperedges. If a dict, must contain (edge_id: label) pairs.
    settings : dict
        Default parameters. Keys that may be useful to override default settings:
        * min_dyad_lw
        * max_dyad_lw
        * dyad_color_cmap
        * edge_fc_cmap
    kwargs : optional keywords
        See `draw_hyperedge_labels` for a description of optional keywords.

    Raises
    ------
    XGIError
        If a SimplicialComplex is passed.

    See Also
    --------
    draw
    draw_xgi_nodes
    draw_xgi_simplices
    draw_node_labels
    draw_hyperedge_labels
    """

    if not max_order:
        max_order = max_edge_order(H)

    dyad_color = _color_arg_to_dict(dyad_color, H.edges, settings["dyad_color_cmap"])
    dyad_lw = _scalar_arg_to_dict(
        dyad_lw, H.edges, settings["min_dyad_lw"], settings["max_dyad_lw"]
    )

    edge_fc = _color_arg_to_dict(edge_fc, H.edges, settings["edge_fc_cmap"])

    # Looping over the hyperedges of different order (reversed) -- nodes will be plotted separately
    for id, he in H.edges.members(dtype=dict).items():
        d = len(he) - 1
        if d > max_order:
            continue
        if d == 1:
            # Drawing the edges
            he = list(he)
            x_coords = [pos[he[0]][0], pos[he[1]][0]]
            y_coords = [pos[he[0]][1], pos[he[1]][1]]
            line = plt.Line2D(
                x_coords,
                y_coords,
                color=dyad_color[id],
                lw=dyad_lw[id],
                zorder=max_order - 1,
            )
            ax.add_line(line)
        else:
            # Hyperedges of order d (d=2: triangles, etc.)
            # Filling the polygon
            coordinates = [[pos[n][0], pos[n][1]] for n in he]
            # Sorting the points counterclockwise (needed to have the correct filling)
            sorted_coordinates = _CCW_sort(coordinates)
            obj = plt.Polygon(
                sorted_coordinates,
                facecolor=edge_fc[id],
                alpha=0.4,
                zorder=max_order - d,
            )
            ax.add_patch(obj)

    if hyperedge_labels:
        # Get all valid keywords by inspecting the signatures of draw_node_labels
        valid_label_kwds = signature(draw_hyperedge_labels).parameters.keys()
        # Remove the arguments of this function (draw_networkx)
        valid_label_kwds = valid_label_kwds - {"H", "pos", "ax", "hyperedge_labels"}
        if any([k not in valid_label_kwds for k in kwargs]):
            invalid_args = ", ".join([k for k in kwargs if k not in valid_label_kwds])
            raise ValueError(f"Received invalid argument(s): {invalid_args}")
        label_kwds = {k: v for k, v in kwargs.items() if k in valid_label_kwds}
        draw_hyperedge_labels(H, pos, hyperedge_labels, ax_edges=ax, **label_kwds)


def draw_xgi_simplices(
    SC,
    pos,
    ax,
    dyad_color,
    dyad_lw,
    edge_fc,
    max_order,
    settings,
    hyperedge_labels,
    **kwargs,
):
    """Draw maximal simplices and pairwise faces.

    Parameters
    ----------
    SC : SimplicialComplex
    ax : matplotlib.pyplot.axes
    pos : dict
        Dictionary of positions node_id:(x,y) of every node.
    dyad_color : str, dict, iterable, or EdgeStat
        Color of the dyadic links.  If str, use the same color for all edges. If a dict, must
        contain (edge_id: color_str) pairs.  If iterable, assume the colors are
        specified in the same order as the edges are found in H.edges. If EdgeStat, use a colormap
        (specified with dyad_color_cmap) associated to it.
    dyad_lw : int, float, dict, iterable, or EdgeStat
        Line width of edges of order 1 (dyadic links).  If int or float, use the same width for all edges.
        If a dict, must contain (edge_id: width) pairs.  If iterable, assume the widths are
        specified in the same order as the edges are found in H.edges. If EdgeStat, use a monotonic
        linear interpolation defined between min_dyad_lw and max_dyad_lw.
    edge_fc : str, dict, iterable, or EdgeStat (default=None)
        Color of the hyperedges.  If str, use the same color for all nodes.  If a dict, must
        contain (edge_id: color_str) pairs.  If other iterable, assume the colors are
        specified in the same order as the hyperedges are found in H.edges. If EdgeStat,
        use the colormap specified with edge_fc_cmap.
    max_order : int
        Maximum of hyperedges to plot.
    hyperedge_labels : bool, or dict
        If True, draw ids on the hyperedges. If a dict, must contain (edge_id: label) pairs.
        Note, we plot only the maximal simplices so if you pass a dict be careful to match its keys
        with the new edge ids in the converted SimplicialComplex. These may differ from the
        edge ids in the given SC.
    settings : dict
        Default parameters. Keys that may be useful to override default settings:
        * min_dyad_lw
        * max_dyad_lw
        * dyad_color_cmap
        * edge_fc_cmap
    kwargs : optional keywords
        See `draw_hyperedge_labels` for a description of optional keywords.


    Raises
    ------
    XGIError
        If a Hypergraph is passed.

    See Also
    --------
    draw
    draw_xgi_nodes
    draw_xgi_hyperedges
    draw_node_labels
    draw_hyperedge_labels
    """

    if max_order:
        max_edges = SC.edges.filterby("order", max_order, "leq").members()
        SC = SimplicialComplex(max_edges)  # SC without simplices larger than max_order

    # Plot only the maximal simplices, thus let's convert the SC to H
    H_ = convert.from_simplicial_complex_to_hypergraph(SC)

    if not max_order:
        max_order = max_edge_order(H_)

    dyad_color = _color_arg_to_dict(dyad_color, H_.edges, settings["dyad_color_cmap"])
    dyad_lw = _scalar_arg_to_dict(
        dyad_lw,
        H_.edges,
        settings["min_dyad_lw"],
        settings["max_dyad_lw"],
    )

    edge_fc = _color_arg_to_dict(edge_fc, H_.edges, settings["edge_fc_cmap"])

    # Looping over the hyperedges of different order (reversed) -- nodes will be plotted separately
    for id, he in H_.edges.members(dtype=dict).items():
        d = len(he) - 1

        if d == 1:
            # Drawing the edges
            he = list(he)
            x_coords = [pos[he[0]][0], pos[he[1]][0]]
            y_coords = [pos[he[0]][1], pos[he[1]][1]]

            line = plt.Line2D(
                x_coords,
                y_coords,
                color=dyad_color[id],
                lw=dyad_lw[id],
                zorder=max_order - 1,
            )
            ax.add_line(line)
        else:
            # Hyperedges of order d (d=2: triangles, etc.)
            # Filling the polygon
            coordinates = [[pos[n][0], pos[n][1]] for n in he]
            # Sorting the points counterclockwise (needed to have the correct filling)
            sorted_coordinates = _CCW_sort(coordinates)
            obj = plt.Polygon(sorted_coordinates, facecolor=edge_fc[id], alpha=0.4)
            ax.add_patch(obj)
            # Drawing all the edges within
            for i, j in combinations(sorted_coordinates, 2):
                x_coords = [i[0], j[0]]
                y_coords = [i[1], j[1]]
                line = plt.Line2D(
                    x_coords, y_coords, color=dyad_color[id], lw=dyad_lw[id]
                )
                ax.add_line(line)

    if hyperedge_labels:
        # Get all valid keywords by inspecting the signatures of draw_node_labels
        valid_label_kwds = signature(draw_hyperedge_labels).parameters.keys()
        # Remove the arguments of this function (draw_networkx)
        valid_label_kwds = valid_label_kwds - {"H", "pos", "ax", "hyperedge_labels"}
        if any([k not in valid_label_kwds for k in kwargs]):
            invalid_args = ", ".join([k for k in kwargs if k not in valid_label_kwds])
            raise ValueError(f"Received invalid argument(s): {invalid_args}")
        label_kwds = {k: v for k, v in kwargs.items() if k in valid_label_kwds}
        draw_hyperedge_labels(H_, pos, hyperedge_labels, ax_edges=ax, **label_kwds)


def _scalar_arg_to_dict(arg, ids, min_val, max_val):
    """Map different types of arguments for drawing style to a dict with scalar values.

    Parameters
    ----------
    arg : int, float, dict, iterable, or NodeStat/EdgeStat
        Attributes for drawing parameter.
    ids : NodeView or EdgeView
        This is the node or edge IDs that attributes get mapped to.
    min_val : int or float
        The minimum value of the drawing parameter.
    max_val : int or float
        The maximum value of the drawing parameter.

    Returns
    -------
    dict
        An ID: attribute dictionary.

    Raises
    ------
    TypeError
        If a int, float, list, dict, or NodeStat/EdgeStat is not passed.
    """
    if isinstance(arg, dict):
        return {id: arg[id] for id in arg if id in ids}
    elif type(arg) in [int, float]:
        return {id: arg for id in ids}
    elif isinstance(arg, NodeStat) or isinstance(arg, EdgeStat):
        vals = np.interp(arg.asnumpy(), [arg.min(), arg.max()], [min_val, max_val])
        return dict(zip(ids, vals))
    elif isinstance(arg, Iterable):
        return {id: arg[idx] for idx, id in enumerate(ids)}
    else:
        raise TypeError(
            f"Argument must be int, float, dict, iterable, or NodeStat/EdgeStat. Received {type(arg)}"
        )


def _color_arg_to_dict(arg, ids, cmap):
    """Map different types of arguments for drawing style to a dict with color values.

    Parameters
    ----------
    arg : str, dict, iterable, or NodeStat/EdgeStat
        Attributes for drawing parameter.
    ids : NodeView or EdgeView
        This is the node or edge IDs that attributes get mapped to.
    cmap : ListedColormap or LinearSegmentedColormap
        colormap to use for NodeStat/EdgeStat.

    Returns
    -------
    dict
        An ID: attribute dictionary.

    Raises
    ------
    TypeError
        If a string, dict, iterable, or NodeStat/EdgeStat is not passed.
    """
    if isinstance(arg, dict):
        return {id: arg[id] for id in arg if id in ids}
    elif isinstance(arg, str):
        return {id: arg for id in ids}
    elif isinstance(arg, NodeStat) or isinstance(arg, EdgeStat):
        if isinstance(cmap, ListedColormap):
            vals = np.interp(arg.asnumpy(), [arg.min(), arg.max()], [0, cmap.N])
        elif isinstance(cmap, LinearSegmentedColormap):
            vals = np.interp(arg.asnumpy(), [arg.min(), arg.max()], [0.1, 0.9])
        else:
            raise XGIError("Invalid colormap!")

        return {id: np.array(cmap(vals[i])).reshape(1, -1) for i, id in enumerate(ids)}
    elif isinstance(arg, Iterable):
        return {id: arg[idx] for idx, id in enumerate(ids)}
    else:
        raise TypeError(
            f"Argument must be str, dict, iterable, or NodeStat/EdgeStat. Received {type(arg)}"
        )


def _CCW_sort(p):
    """
    Sort the input 2D points counterclockwise.
    """
    p = np.array(p)
    mean = np.mean(p, axis=0)
    d = p - mean
    s = np.arctan2(d[:, 0], d[:, 1])
    return p[np.argsort(s), :]


def draw_node_labels(
    H,
    pos,
    node_labels=False,
    font_size_nodes=10,
    font_color_nodes="black",
    font_family_nodes="sans-serif",
    font_weight_nodes="normal",
    alpha_nodes=None,
    bbox_nodes=None,
    horizontalalignment_nodes="center",
    verticalalignment_nodes="center",
    ax_nodes=None,
    clip_on_nodes=True,
):
    """Draw node labels on the hypergraph or simplicial complex.

    Parameters
    ----------
    H : Hypergraph or SimplicialComplex.
    pos : dict
        Dictionary of positions node_id:(x,y).
    node_labels : bool, or dict (default=False)
        If True, draw ids on the nodes. If a dict, must contain (node_id: label) pairs.
    font_size_nodes : int (default=10)
        Font size for text labels.
    font_color_nodes : str (default='black')
        Font color string.
    font_family_nodes : str (default='sans-serif')
        Font family.
    font_weight_nodes : str (default='normal')
        Font weight.
    alpha_nodes : float (default=None)
        The text transparency.
    bbox_nodes : Matplotlib bbox (default is Matplotlib's ax.text default)
        Specify text box properties (e.g. shape, color etc.) for node labels.
    horizontalalignment_nodes : str (default='center')
        Horizontal alignment {'center', 'right', 'left'}.
    verticalalignment_nodes : str (default='center')
        Vertical alignment {'center', 'top', 'bottom', 'baseline', 'center_baseline'}.
    ax_nodes : matplotlib.pyplot.axes (default=None)
        Draw the graph in the specified Matplotlib axes.
    clip_on_nodes : bool (default=True)
        Turn on clipping of node labels at axis boundaries.

    Returns
    -------
    dict
        `dict` of labels keyed by node id.

    See Also
    --------
    draw
    draw_xgi_nodes
    draw_xgi_hyperedges
    draw_xgi_simplices
    draw_hyperedge_labels
    """
    if ax_nodes is None:
        ax = plt.gca()
    else:
        ax = ax_nodes

    if node_labels is True:
        node_labels = {id: id for id in H.nodes}

    # Plot the labels in the last layer
    zorder = max_edge_order(H) + 1

    text_items = {}
    for id, label in node_labels.items():
        (x, y) = pos[id]

        if not isinstance(label, str):
            label = str(label)

        t = ax.text(
            x,
            y,
            label,
            size=font_size_nodes,
            color=font_color_nodes,
            family=font_family_nodes,
            weight=font_weight_nodes,
            alpha=alpha_nodes,
            horizontalalignment=horizontalalignment_nodes,
            verticalalignment=verticalalignment_nodes,
            transform=ax.transData,
            bbox=bbox_nodes,
            clip_on=clip_on_nodes,
            zorder=zorder,
        )
        text_items[id] = t

    return text_items


def draw_hyperedge_labels(
    H,
    pos,
    hyperedge_labels=False,
    font_size_edges=10,
    font_color_edges="black",
    font_family_edges="sans-serif",
    font_weight_edges="normal",
    alpha_edges=None,
    bbox_edges=None,
    horizontalalignment_edges="center",
    verticalalignment_edges="center",
    ax_edges=None,
    rotate_edges=False,
    clip_on_edges=True,
):
    """Draw hyperedge labels on the hypegraph or simplicial complex.

    Parameters
    ----------
    H : Hypergraph.
    pos : dict
        Dictionary of positions node_id:(x,y).
    hyperedge_labels : bool, or dict (default=False)
        If True, draw ids on the hyperedges. If a dict, must contain (edge_id: label) pairs.
    font_size_edges : int (default=10)
        Font size for text labels.
    font_color_edges : str (default='black')
        Font color string.
    font_family_edges : str (default='sans-serif')
        Font family.
    font_weight_edges : str (default='normal')
        Font weight.
    alpha_edges : float (default=None)
        The text transparency.
    bbox_edges : Matplotlib bbox (default={boxstyle='round', ec=(1.0, 1.0, 1.0), fc=(1.0, 1.0, 1.0)})
        Specify text box properties (e.g. shape, color etc.) for edge labels.
    horizontalalignment_edges : str (default='center')
        Horizontal alignment {'center', 'right', 'left'}.
    verticalalignment_edges: str (default='center')
        Vertical alignment {'center', 'top', 'bottom', 'baseline', 'center_baseline'}.
    ax_edges : matplotlib.pyplot.axes (default=None)
        Draw the graph in the specified Matplotlib axes.
    rotate_edges : bool (default=False)
        Rotate edge labels for dyadic links to lie parallel to edges.
    clip_on_edges: bool (default=True)
        Turn on clipping of hyperedge labels at axis boundaries.

    Returns
    -------
    dict
        `dict` of labels keyed by hyperedge id.

    See Also
    --------
    draw
    draw_xgi_nodes
    draw_xgi_hyperedges
    draw_xgi_simplices
    draw_node_labels
    """
    if ax_edges is None:
        ax = plt.gca()
    else:
        ax = ax_edges

    if hyperedge_labels is True:
        hyperedge_labels = {id: id for id in H.edges}

    text_items = {}
    for id, label in hyperedge_labels.items():
        he = H.edges.members(id)
        coordinates = [[pos[n][0], pos[n][1]] for n in he]
        x, y = np.mean(coordinates, axis=0)

        if len(he) == 2:
            # Rotate edge labels for dyadic links to lie parallel to edges
            if rotate_edges:
                x_diff, y_diff = np.subtract(coordinates[1], coordinates[0])
                angle = np.arctan2(y_diff, x_diff) / (2.0 * np.pi) * 360
                # Make label orientation "right-side-up"
                if angle > 90:
                    angle -= 180
                if angle < -90:
                    angle += 180
                # Transform data coordinate angle to screen coordinate angle
                xy = np.array((x, y))
                trans_angle = ax.transData.transform_angles(
                    np.array((angle,)), xy.reshape((1, 2))
                )[0]
            else:
                trans_angle = 0.0
        else:
            trans_angle = 0.0

        # Use default box of white with white border
        if bbox_edges is None:
            bbox = dict(boxstyle="round", ec=(1.0, 1.0, 1.0), fc=(1.0, 1.0, 1.0))
        else:
            bbox = bbox_edges

        if not isinstance(label, str):
            label = str(label)

        t = ax.text(
            x,
            y,
            label,
            size=font_size_edges,
            color=font_color_edges,
            family=font_family_edges,
            weight=font_weight_edges,
            alpha=alpha_edges,
            horizontalalignment=horizontalalignment_edges,
            verticalalignment=verticalalignment_edges,
            rotation=trans_angle,
            transform=ax.transData,
            bbox=bbox,
            clip_on=clip_on_edges,
        )
        text_items[id] = t

    return text_items



"""Generate random (non-uniform) hypergraphs."""
import math
import random
import warnings
from collections import defaultdict
from itertools import combinations

import networkx as nx
import numpy as np
from scipy.special import comb

from ..classes import SimplicialComplex
from .classic import empty_hypergraph, flag_complex_d2, ring_lattice

__all__ = [
    "chung_lu_hypergraph",
    "dcsbm_hypergraph",
    "random_hypergraph",
    "random_simplicial_complex",
    "random_flag_complex_d2",
    "random_flag_complex",
    "watts_strogatz_hypergraph",
]


def chung_lu_hypergraph(k1, k2, seed=None):
    """A function to generate a Chung-Lu hypergraph

    Parameters
    ----------
    k1 : dictionary
        Dictionary where the keys are node ids
        and the values are node degrees.
    k2 : dictionary
        Dictionary where the keys are edge ids
        and the values are edge sizes.
    seed : integer or None (default)
            The seed for the random number generator.

    Returns
    -------
    Hypergraph object
        The generated hypergraph

    Warns
    -----
    warnings.warn
        If the sums of the edge sizes and node degrees are not equal, the
        algorithm still runs, but raises a warning.

    Notes
    -----
    The sums of k1 and k2 should be the same. If they are not the same,
    this function returns a warning but still runs.

    References
    ----------
    Implemented by Mirah Shi in HyperNetX and described for
    bipartite networks by Aksoy et al. in https://doi.org/10.1093/comnet/cnx001

    Example
    -------
    >>> import xgi
    >>> import random
    >>> n = 100
    >>> k1 = {i : random.randint(1, 100) for i in range(n)}
    >>> k2 = {i : sorted(k1.values())[i] for i in range(n)}
    >>> H = xgi.chung_lu_hypergraph(k1, k2)

    """
    if seed is not None:
        random.seed(seed)

    # sort dictionary by degree in decreasing order
    node_labels = [n for n, _ in sorted(k1.items(), key=lambda d: d[1], reverse=True)]
    edge_labels = [m for m, _ in sorted(k2.items(), key=lambda d: d[1], reverse=True)]

    m = len(k2)

    if sum(k1.values()) != sum(k2.values()):
        warnings.warn(
            "The sum of the degree sequence does not match the sum of the size sequence"
        )

    S = sum(k1.values())

    H = empty_hypergraph()
    H.add_nodes_from(node_labels)

    for u in node_labels:
        j = 0
        v = edge_labels[j]  # start from beginning every time
        p = min((k1[u] * k2[v]) / S, 1)

        while j < m:
            if p != 1:
                r = random.random()
                try:
                    j = j + math.floor(math.log(r) / math.log(1 - p))
                except ZeroDivisionError:
                    j = np.inf

            if j < m:
                v = edge_labels[j]
                q = min((k1[u] * k2[v]) / S, 1)
                r = random.random()
                if r < q / p:
                    # no duplicates
                    H.add_node_to_edge(v, u)
                p = q
                j = j + 1

    return H


def dcsbm_hypergraph(k1, k2, g1, g2, omega, seed=None):
    """A function to generate a DCSBM hypergraph.

    Parameters
    ----------
    k1 : dict
        This is a dictionary where the keys are node ids
        and the values are node degrees.
    k2 : dict
        This is a dictionary where the keys are edge ids
        and the values are edge sizes.
    g1 : dict
        This a dictionary where the keys are node ids
        and the values are the group ids to which the node belongs.
        The keys must match the keys of k1.
    g2 : dict
        This a dictionary where the keys are edge ids
        and the values are the group ids to which the edge belongs.
        The keys must match the keys of k2.
    omega : 2D numpy array
        This is a matrix with entries which specify the number of edges
        between a given node community and edge community.
        The number of rows must match the number of node communities
        and the number of columns must match the number of edge
        communities.
    seed : int or None (default)
        Seed for the random number generator.

    Returns
    -------
    Hypergraph

    Warns
    -----
    warnings.warn
        If the sums of the edge sizes and node degrees are not equal, the
        algorithm still runs, but raises a warning.
        Also if the sum of the omega matrix does not match the sum of degrees,
        a warning is raised.

    Notes
    -----
    The sums of k1 and k2 should be the same. If they are not the same, this function
    returns a warning but still runs. The sum of k1 (and k2) and omega should be the
    same. If they are not the same, this function returns a warning but still runs and
    the number of entries in the incidence matrix is determined by the omega matrix.

    References
    ----------
    Implemented by Mirah Shi in HyperNetX and described for bipartite networks by
    Larremore et al. in https://doi.org/10.1103/PhysRevE.90.012805

    Examples
    --------
    >>> import xgi; import random; import numpy as np
    >>> n = 50
    >>> k1 = {i : random.randint(1, n) for i in range(n)}
    >>> k2 = {i : sorted(k1.values())[i] for i in range(n)}
    >>> g1 = {i : random.choice([0, 1]) for i in range(n)}
    >>> g2 = {i : random.choice([0, 1]) for i in range(n)}
    >>> omega = np.array([[n//2, 10], [10, n//2]])
    >>> # H = xgi.dcsbm_hypergraph(k1, k2, g1, g2, omega)

    """
    if seed is not None:
        random.seed(seed)

    # sort dictionary by degree in decreasing order
    node_labels = [n for n, _ in sorted(k1.items(), key=lambda d: d[1], reverse=True)]
    edge_labels = [m for m, _ in sorted(k2.items(), key=lambda d: d[1], reverse=True)]

    # these checks verify that the sum of node and edge degrees and the sum of node degrees
    # and the sum of community connection matrix differ by less than a single edge.
    if abs(sum(k1.values()) - sum(k2.values())) > 1:
        warnings.warn(
            "The sum of the degree sequence does not match the sum of the size sequence"
        )

    if abs(sum(k1.values()) - np.sum(omega)) > 1:
        warnings.warn(
            "The sum of the degree sequence does not match the entries in the omega matrix"
        )

    # get indices for each community
    community1_nodes = defaultdict(list)
    for label in node_labels:
        group = g1[label]
        community1_nodes[group].append(label)

    community2_nodes = defaultdict(list)
    for label in edge_labels:
        group = g2[label]
        community2_nodes[group].append(label)

    H = empty_hypergraph()
    H.add_nodes_from(node_labels)

    kappa1 = defaultdict(lambda: 0)
    kappa2 = defaultdict(lambda: 0)
    for id, g in g1.items():
        kappa1[g] += k1[id]
    for id, g in g2.items():
        kappa2[g] += k2[id]

    for group1 in community1_nodes.keys():
        for group2 in community2_nodes.keys():
            # for each constant probability patch
            try:
                group_constant = omega[group1, group2] / (
                    kappa1[group1] * kappa2[group2]
                )
            except ZeroDivisionError:
                group_constant = 0

            for u in community1_nodes[group1]:
                j = 0
                v = community2_nodes[group2][j]  # start from beginning every time
                # max probability
                p = min(k1[u] * k2[v] * group_constant, 1)
                while j < len(community2_nodes[group2]):
                    if p != 1:
                        r = random.random()
                        try:
                            j = j + math.floor(math.log(r) / math.log(1 - p))
                        except ZeroDivisionError:
                            j = np.inf
                    if j < len(community2_nodes[group2]):
                        v = community2_nodes[group2][j]
                        q = min((k1[u] * k2[v]) * group_constant, 1)
                        r = random.random()
                        if r < q / p:
                            # no duplicates
                            H.add_node_to_edge(v, u)
                        p = q
                        j = j + 1
    return H


def random_hypergraph(N, ps, seed=None):
    """Generates a random hypergraph

    Generate N nodes, and connect any d+1 nodes
    by a hyperedge with probability ps[d-1].

    Parameters
    ----------
    N : int
        Number of nodes
    ps : list of float
        List of probabilities (between 0 and 1) to create a
        hyperedge at each order d between any d+1 nodes. For example,
        ps[0] is the wiring probability of any edge (2 nodes), ps[1]
        of any triangles (3 nodes).
    seed : integer or None (default)
            Seed for the random number generator.

    Returns
    -------
    Hypergraph object
        The generated hypergraph

    References
    ----------
    Described as 'random hypergraph' by M. Dewar et al. in https://arxiv.org/abs/1703.07686

    Example
    -------
    >>> import xgi
    >>> H = xgi.random_hypergraph(50, [0.1, 0.01])

    """
    if seed is not None:
        np.random.seed(seed)

    if (np.any(np.array(ps) < 0)) or (np.any(np.array(ps) > 1)):
        raise ValueError("All elements of ps must be between 0 and 1 included.")

    nodes = range(N)
    hyperedges = []

    for i, p in enumerate(ps):
        d = i + 1  # order, ps[0] is prob of edges (d=1)

        potential_edges = combinations(nodes, d + 1)
        n_comb = comb(N, d + 1, exact=True)
        mask = np.random.random(size=n_comb) <= p  # True if edge to keep

        edges_to_add = [e for e, val in zip(potential_edges, mask) if val]

        hyperedges += edges_to_add

    H = empty_hypergraph()
    H.add_nodes_from(nodes)
    H.add_edges_from(hyperedges)

    return H


def random_simplicial_complex(N, ps, seed=None):
    """Generates a random hypergraph

    Generate N nodes, and connect any d+1 nodes
    by a simplex with probability ps[d-1]. For each simplex,
    add all its subfaces if they do not already exist.

    Parameters
    ----------
    N : int
        Number of nodes
    ps : list of float
        List of probabilities (between 0 and 1) to create a
        hyperedge at each order d between any d+1 nodes. For example,
        ps[0] is the wiring probability of any edge (2 nodes), ps[1]
        of any triangles (3 nodes).
    seed : int or None (default)
        The seed for the random number generator

    Returns
    -------
    Simplicialcomplex object
        The generated simplicial complex

    References
    ----------
    Described as 'random simplicial complex' in
    "Simplicial Models of Social Contagion", Nature Communications 10(1), 2485,
    by I. Iacopini, G. Petri, A. Barrat & V. Latora (2019).
    https://doi.org/10.1038/s41467-019-10431-6

    Example
    -------
    >>> import xgi
    >>> H = xgi.random_simplicial_complex(20, [0.1, 0.01])

    """

    if seed is not None:
        np.random.seed(seed)

    if (np.any(np.array(ps) < 0)) or (np.any(np.array(ps) > 1)):
        raise ValueError("All elements of ps must be between 0 and 1 included.")

    nodes = range(N)
    simplices = []

    for i, p in enumerate(ps):
        d = i + 1  # order, ps[0] is prob of edges (d=1)

        potential_simplices = combinations(nodes, d + 1)
        n_comb = comb(N, d + 1, exact=True)
        mask = np.random.random(size=n_comb) <= p  # True if simplex to keep

        simplices_to_add = [e for e, val in zip(potential_simplices, mask) if val]

        simplices += simplices_to_add

    S = SimplicialComplex()
    S.add_nodes_from(nodes)
    S.add_simplices_from(simplices)

    return S


def random_flag_complex_d2(N, p, seed=None):
    """Generate a maximal simplicial complex (up to order 2) from a
    :math:`G_{N,p}` Erdős-Rényi random graph by filling all empty triangles with 2-simplices.

    Parameters
    ----------
    N : int
        Number of nodes
    p : float
        Probabilities (between 0 and 1) to create an edge
        between any 2 nodes
    seed : int or None (default)
        The seed for the random number generator

    Returns
    -------
    SimplicialComplex

    Notes
    -----
    Computing all cliques quickly becomes heavy for large networks.
    """
    if seed is not None:
        random.seed(seed)

    if (p < 0) or (p > 1):
        raise ValueError("p must be between 0 and 1 included.")

    G = nx.fast_gnp_random_graph(N, p, seed=seed)

    return flag_complex_d2(G)


def random_flag_complex(N, p, max_order=2, seed=None):
    """Generate a flag (or clique) complex from a
    :math:`G_{N,p}` Erdős-Rényi random graph by filling all cliques up to dimension max_order.

    Parameters
    ----------
    N : int
        Number of nodes
    p : float
        Probabilities (between 0 and 1) to create an edge
        between any 2 nodes
    max_order : int
        maximal dimension of simplices to add to the output simplicial complex
    seed : int or None (default)
        The seed for the random number generator

    Returns
    -------
    SimplicialComplex

    Notes
    -----
    Computing all cliques quickly becomes heavy for large networks.
    """

    if (p < 0) or (p > 1):
        raise ValueError("p must be between 0 and 1 included.")

    G = nx.fast_gnp_random_graph(N, p, seed=seed)

    nodes = G.nodes()
    edges = list(G.edges())

    # compute all triangles to fill
    max_cliques = list(nx.find_cliques(G))

    S = SimplicialComplex()
    S.add_nodes_from(nodes)
    S.add_simplices_from(max_cliques, max_order=max_order)

    return S


def watts_strogatz_hypergraph(n, d, k, l, p, seed=None):
    if seed is not None:
        np.random.seed(seed)
    H = ring_lattice(n, d, k, l)
    to_remove = []
    to_add = []
    for e in H.edges:
        if np.random.random() < p:
            to_remove.append(e)
            node = min(H.edges.members(e))
            neighbors = np.random.choice(H.nodes, size=d - 1)
            to_add.append(np.append(neighbors, node))
    H.remove_edges_from(to_remove)
    H.add_edges_from(to_add)
    return H



"""Algorithms for finding the degree assortativity of a hypergraph."""
import random
from itertools import combinations

import numpy as np

from ..classes import is_uniform, unique_edge_sizes
from ..exception import XGIError

__all__ = ["dynamical_assortativity", "degree_assortativity"]


def dynamical_assortativity(H):
    """Computes the dynamical assortativity of a uniform hypergraph.

    Parameters
    ----------
    H : xgi.Hypergraph
        Hypergraph of interest

    Returns
    -------
    float
        The dynamical assortativity

    Raises
    ------
    XGIError
        If the hypergraph is not uniform, or if there are no nodes
        or no edges

    References
    ----------
    Nicholas Landry and Juan G. Restrepo,
    Hypergraph assortativity: A dynamical systems perspective,
    Chaos 2022.
    DOI: 10.1063/5.0086905

    """
    if H.num_nodes == 0 or H.num_edges == 0:
        raise XGIError("Hypergraph must contain nodes and edges!")

    if not is_uniform(H):
        raise XGIError("Hypergraph must be uniform!")

    if 1 in unique_edge_sizes(H):
        raise XGIError("No singleton edges!")

    d = H.nodes.degree
    degs = d.asdict()
    k1 = d.mean()
    k2 = d.moment(2)
    kk1 = np.mean(
        [
            degs[n1] * degs[n2]
            for e in H.edges
            for n1, n2 in combinations(H.edges.members(e), 2)
        ]
    )

    return kk1 * k1**2 / k2**2 - 1


def degree_assortativity(H, kind="uniform", exact=False, num_samples=1000):
    """Computes the degree assortativity of a hypergraph

    Parameters
    ----------
    H : Hypergraph
        The hypergraph of interest
    kind : str, default: "uniform"
        the type of degree assortativity. valid choices are
        "uniform", "top-2", and "top-bottom".
    exact : bool, default: False
        whether to compute over all edges or
        sample randomly from the set of edges
    num_samples : int, default: 1000
        if not exact, specify the number of samples for the computation.

    Returns
    -------
    float
        the degree assortativity

    References
    ----------
    Phil Chodrow,
    Configuration models of random hypergraphs,
    Journal of Complex Networks 2020.
    DOI: 10.1093/comnet/cnaa018
    """
    degs = H.degree()
    if exact:
        k1k2 = [
            choose_degrees(H.edges.members(e), degs, kind)
            for e in H.edges
            if len(H.edges.members(e)) > 1
        ]
    else:
        edges = [e for e in H.edges if len(H.edges.members(e)) > 1]
        k1k2 = [
            choose_degrees(H.edges.members(random.choice(edges)), degs, kind)
            for _ in range(num_samples)
        ]

    rho = np.corrcoef(np.array(k1k2).T)[0, 1]
    if np.isnan(rho):
        return 0
    return rho


def choose_degrees(e, k, kind="uniform"):
    """Choose the degrees of two nodes in a hyperedge.

    Parameters
    ----------
    e : iterable
        the members in a hyperedge
    k : dict
        the degrees where keys are node IDs and values are degrees
    kind : str, default: "uniform"
        the type of degree assortativity, options are
        "uniform", "top-2", and "top-bottom".

    Returns
    -------
    tuple
        two degrees selected from the edge

    Raises
    ------
    XGIError
        if invalid assortativity function chosen

    References
    ----------
    Phil Chodrow,
    Configuration models of random hypergraphs,
    Journal of Complex Networks 2020.
    DOI: 10.1093/comnet/cnaa018
    """
    e = list(e)
    if len(e) > 1:
        if kind == "uniform":
            i = np.random.randint(len(e))
            j = i
            while i == j:
                j = np.random.randint(len(e))
            return (k[e[i]], k[e[j]])

        elif kind == "top-2":
            degs = sorted([k[i] for i in e])[-2:]
            random.shuffle(degs)
            return degs

        elif kind == "top-bottom":
            # this selects the largest and smallest degrees in one line
            degs = sorted([k[i] for i in e])[:: len(e) - 1]
            random.shuffle(degs)
            return degs

        else:
            raise XGIError("Invalid choice function!")
    else:
        raise XGIError("Edge must have more than one member!")



"""Base class for undirected simplicial complexes.

The SimplicialComplex class allows any hashable object as a node
and can associate key/value attribute pairs with each undirected simplex and node.

Multi-simplices are not allowed.
"""

from collections.abc import Hashable, Iterable
from itertools import combinations, count
from warnings import warn

from ..exception import XGIError
from ..utils.utilities import powerset, update_uid_counter
from .hypergraph import Hypergraph
from .reportviews import EdgeView, NodeView

__all__ = ["SimplicialComplex"]


class SimplicialComplex(Hypergraph):
    r"""A class to represent undirected simplicial complexes.

    A simplicial complex is a collection of subsets of a set of *nodes* or *vertices*.
    It is a pair :math:`(V, E)`, where :math:`V` is a set of elements called
    *nodes* or *vertices*, and :math:`E` is a set whose elements are subsets of
    :math:`V`, that is, each :math:`e \in E` satisfies :math:`e \subset V`.  The
    elements of :math:`E` are called *simplices*. Additionally, if a simplex is part of
    a simplicial complex, all its faces must be too. This makes simplicial complexes
    a special case of hypergraphs.

    The SimplicialComplex class allows any hashable object as a node and can associate
    attributes to each node, simplex, or the simplicial complex itself, in the form of key/value
    pairs.


    Parameters
    ----------
    incoming_data : input simplicial complex data (optional, default: None)
        Data to initialize the simplicial complex. If None (default), an empty
        simplicial complex is created, i.e. one with no nodes or simplices.
        The data can be in the following formats:

        * simplex list
        * simplex dictionary
        * 2-column Pandas dataframe (bipartite edges)
        * Scipy/Numpy incidence matrix
        * SimplicialComplex object.

    **attr : dict, optional, default: None
        Attributes to add to the simplicial complex as key, value pairs.

    Notes
    -----
    Unique IDs are assigned to each node and simplex internally and are used to refer to
    them throughout.

    The `attr` keyword arguments are added as simplicial complex attributes. To add node
    or simplex attributes see :meth:`add_node` and :meth:`add_simplex`. Methods such as
    :meth:`add_simplex` replace :class:`Hypergraph` methods such as :meth:`add_edge`
    which here raise an error.

    Examples
    --------
    >>> import xgi
    >>> S = xgi.SimplicialComplex([[1, 2, 3], [4], [5, 6], [6, 7, 8]])
    >>> S.nodes
    NodeView((1, 2, 3, 4, 5, 6, 7, 8))
    >>> S.edges
    EdgeView((0, 1, 2, 3, 4, 5, 6, 7, 8, 9))

    """

    def __init__(self, incoming_data=None, **attr):
        self._edge_uid = count()
        self._hypergraph = self._hypergraph_attr_dict_factory()
        self._node = self._node_dict_factory()
        self._node_attr = self._node_attr_dict_factory()
        self._edge = self._hyperedge_dict_factory()
        self._edge_attr = self._hyperedge_attr_dict_factory()

        self._nodeview = NodeView(self)
        self._edgeview = EdgeView(self)

        if incoming_data is not None:
            from ..convert import convert_to_simplicial_complex

            convert_to_simplicial_complex(incoming_data, create_using=self)
        self._hypergraph.update(attr)  # must be after convert

    def __str__(self):
        """Returns a short summary of the simplicial complex.

        Returns
        -------
        string
            SimplicialComplex information

        Examples
        --------
        >>> import xgi
        >>> H = xgi.SimplicialComplex(name="foo")
        >>> str(H)
        "SimplicialComplex named 'foo' with 0 nodes and 0 simplices"

        """
        try:
            return f"{type(self).__name__} named '{self['name']}' with {self.num_nodes} nodes and {self.num_edges} simplices"
        except XGIError:
            return f"Unnamed {type(self).__name__} with {self.num_nodes} nodes and {self.num_edges} simplices"

    def add_edge(self, edge, id=None, **attr):
        """add_edge is deprecated in SimplicialComplex. Use add_simplex instead"""
        warn("add_edge is deprecated in SimplicialComplex. Use add_simplex instead")
        return self.add_simplex(edge, id=None, **attr)

    def add_edges_from(self, ebunch_to_add, max_order=None, **attr):
        """add_edges_from is deprecated in SimplicialComplex. Use add_simplices_from instead"""
        warn(
            "add_edges_from is deprecated in SimplicialComplex. Use add_simplices_from instead"
        )
        return self.add_simplices_from(ebunch_to_add, max_order=None, **attr)

    def add_weighted_edges_from(
        self, ebunch_to_add, max_order=None, weight="weight", **attr
    ):
        """add_weighted_edges_from is deprecated in SimplicialComplex. Use add_weighted_simplices_from instead"""
        warn(
            "add_weighted_edges_from is deprecated in SimplicialComplex. Use add_weighted_simplices_from instead"
        )
        return self.add_weighted_simplices_from(
            ebunch_to_add, max_order=None, weight="weight", **attr
        )

    def remove_edge(self, id):
        """remove_edge is deprecated in SimplicialComplex. Use remove_simplex_id instead"""
        warn(
            "remove_edge is deprecated in SimplicialComplex. Use remove_simplex_id instead"
        )
        return self.remove_simplex_id(id, **attr)

    def remove_edges_from(self, ebunch):
        """remove_edges_from is deprecated in SimplicialComplex. Use remove_simplex_ids_from instead"""
        warn(
            "remove_edges_from is deprecated in SimplicialComplex. Use remove_simplex_ids_from instead"
        )
        return self.remove_simplex_ids_from(ebunch)

    def add_node_to_edge(self, edge, node):
        """add_node_to_edge is not implemented in SimplicialComplex."""
        raise XGIError("add_node_to_edge is not implemented in SimplicialComplex.")

    def remove_node(self, n, strong=False):
        """remove_node is not implemented in SimplicialComplex."""
        raise XGIError("remove_node is not implemented in SimplicialComplex.")

    def _add_simplex(self, members, id=None, **attr):
        """Helper function to add a simplex to a simplicial complex, without any
        check. Does not automatically update self._edge_uid"""

        self._edge[id] = set()
        for node in members:
            if node not in self._node:
                if node is None:
                    raise ValueError("None cannot be a node")
                self._node[node] = set()
                self._node_attr[node] = self._node_attr_dict_factory()
            self._node[node].add(id)

        self._edge[id] = members
        self._edge_attr[id] = self._hyperedge_attr_dict_factory()
        self._edge_attr[id].update(attr)

    def _add_face(self, members):
        """Helper function to add a face to a simplicial complex, without any
        check, and without attributes. Automatically updates self._edge_uid"""

        id = next(self._edge_uid)
        self._edge[id] = frozenset(members)

        for n in members:
            if n not in self._node:
                self._node[n] = set()
                self._node_attr[n] = self._node_attr_dict_factory()
            self._node[n].add(id)

        self._edge_attr[id] = self._hyperedge_attr_dict_factory()

    def add_simplex(self, members, id=None, **attr):
        """Add a simplex to the simplicial complex, and all its subfaces that do
        not exist yet.

        Simplex attributes can be specified with keywords or by directly
        accessing the simplex's attribute dictionary. The attributes do not propagate
        to the subfaces.

        Parameters
        ----------
        members : Iterable
            An iterable of the ids of the nodes contained in the new simplex.
        id : hashable, default None
            Id of the new simplex. If None, a unique numeric ID will be created.
        **attr : dict, optional
            Attributes of the new simplex.

        Raises
        -----
        XGIError
            If `members` is empty.

        See Also
        --------
        add_simplices_from : Add a collection of simplices.

        Notes
        -----
        Currently cannot add empty simplices.

        Examples
        --------

        Add simplices with or without specifying an simplex id.

        >>> import xgi
        >>> S = xgi.SimplicialComplex()
        >>> S.add_simplex([1, 2, 3])
        >>> S.edges.members() # doctest: +NORMALIZE_WHITESPACE +SKIP
        [frozenset({1, 2, 3}), frozenset({2, 3}),
            frozenset({1, 2}), frozenset({1, 3})]
        >>> S.add_simplex([3, 4], id='myedge')
        >>> S.edges
        EdgeView((0, 1, 2, 3, 'myedge'))

        Access attributes using square brackets.  By default no attributes are created.

        >>> S.edges[0]
        {}
        >>> S.add_simplex([1, 4], color='red', place='peru')
        >>> S.edges
        EdgeView((0, 1, 2, 3, 'myedge', 4))
        >>> S.edges[4]
        {'color': 'red', 'place': 'peru'}
        """

        try:
            members = frozenset(members)
        except TypeError:
            raise XGIError("The simplex cannot be cast to a frozenset.")

        if not members:
            raise XGIError("Cannot add an empty edge")

        if self.has_simplex(members):
            return

        if id in self._edge.keys():  # check that uid is not present yet
            warn(f"uid {id} already exists, cannot add simplex {members}")
            return

        id = next(self._edge_uid) if not id else id

        self._add_simplex(members, id, **attr)

        # set self._edge_uid correctly
        update_uid_counter(self, id)

        # add all subfaces
        faces = self._subfaces(members)
        faces = set(faces)  # get unique faces
        for members_sub in faces:

            # check that it does not exist yet (based on members, not ID)
            if not members_sub or self.has_simplex(members_sub):
                continue

            self._add_face(members_sub)

    def _subfaces(self, simplex, all=True):
        """Returns list of subfaces of simplex.

        Parameters
        ----------
        simplex: an iterable of hashables
            A list of node ids
        all: bool, default: True
            Whether to return all of the subfaces or just
            those of the order below

        Returns
        -------
        faces: list of iterables
            The list containing the subfaces of the
            given simplex

        """
        size = len(simplex)
        faces = []
        if all:
            for n in range(size, 2, -1):
                for face in combinations(simplex, n - 1):
                    faces.append(face)
        else:
            for face in combinations(simplex, size - 1):
                faces.append(face)
        return faces

    def _supfaces(self, simplex):
        """Returns list of simplices that contain simplex"""

        return [s for s in self._edge.values() if simplex < s]

    def _supfaces_id(self, simplex):
        """Returns list of IDs of simplices that contain simplex"""

        return [id_ for id_, s in self._edge.items() if simplex < s]

    def add_simplices_from(self, ebunch_to_add, max_order=None, **attr):
        """Add multiple edges with optional attributes.

        Parameters
        ----------
        ebunch_to_add : Iterable

            An iterable of simplices.  This may be an iterable of iterables (Format 1),
            where each element contains the members of the simplex specified as valid node IDs.
            Alternatively, each element could also be a tuple in any of the following
            formats:

            * Format 2: 2-tuple (members, simplex_id), or
            * Format 3: 2-tuple (members, attr), or
            * Format 4: 3-tuple (members, simplex_id, attr),

            where `members` is an iterable of node IDs, `simplex_id` is a hashable to use
            as simplex ID, and `attr` is a dict of attributes. Finally, `ebunch_to_add`
            may be a dict of the form `{simplex_id: simplex_members}` (Format 5).

            Formats 2 and 3 are unambiguous because `attr` dicts are not hashable, while `id`s must be.
            In Formats 2-4, each element of `ebunch_to_add` must have the same length,
            i.e. you cannot mix different formats.  The iterables containing simplex
            members cannot be strings.

        attr : \*\*kwargs, optional
            Additional attributes to be assigned to all simplices. Attribues specified via
            `ebunch_to_add` take precedence over `attr`.

        See Also
        --------
        add_simplex : add a single simplex
        add_weighted_simplices_from : convenient way to add weighted simplices

        Notes
        -----
        Adding the same simplex twice will add it only once. Currently
        cannot add empty simplices; the method skips over them.

        Examples
        --------
        >>> import xgi
        >>> S = xgi.SimplicialComplex()

        When specifying simplices by their members only, numeric simplex IDs will be assigned
        automatically.

        >>> S.add_simplices_from([[0, 1], [1, 2], [2, 3, 4]])
        >>> S.edges.members(dtype=dict) # doctest: +SKIP
        {0: frozenset({0, 1}), 1: frozenset({1, 2}), 2: frozenset({2, 3, 4}), 3: frozenset({2, 3}), 4: frozenset({2, 4}), 5: frozenset({3, 4})}

        Custom simplex ids can be specified using a dict.

        >>> S = xgi.SimplicialComplex()
        >>> S.add_simplices_from({'one': [0, 1], 'two': [1, 2], 'three': [2, 3, 4]})
        >>> S.edges.members(dtype=dict) # doctest: +SKIP
        {'one': frozenset({0, 1}), 'two': frozenset({1, 2}), 'three': frozenset({2, 3, 4}), 0: frozenset({2, 3}), 1: frozenset({2, 4}), 2: frozenset({3, 4})}

        You can use the dict format to easily add simplices from another simplicial complex.

        >>> S2 = xgi.SimplicialComplex()
        >>> S2.add_simplices_from(S.edges.members(dtype=dict))
        >>> list(S.edges) == list(S2.edges)
        True

        Alternatively, simplex ids can be specified using an iterable of 2-tuples.

        >>> S = xgi.SimplicialComplex()
        >>> S.add_simplices_from([([0, 1], 'one'), ([1, 2], 'two'), ([2, 3, 4], 'three')])
        >>> S.edges.members(dtype=dict) # doctest: +SKIP
        {'one': frozenset({0, 1}), 'two': frozenset({1, 2}), 'three': frozenset({2, 3, 4}), 0: frozenset({2, 3}), 1: frozenset({2, 4}), 2: frozenset({3, 4})}

        Attributes for each simplex may be specified using a 2-tuple for each simplex.
        Numeric IDs will be assigned automatically.

        >>> S = xgi.SimplicialComplex()
        >>> simplices = [
        ...     ([0, 1], {'color': 'red'}),
        ...     ([1, 2], {'age': 30}),
        ...     ([2, 3, 4], {'color': 'blue', 'age': 40}),
        ... ]
        >>> S.add_simplices_from(simplices)
        >>> {e: S.edges[e] for e in S.edges}
        {0: {'color': 'red'}, 1: {'age': 30}, 2: {'color': 'blue', 'age': 40}, 3: {}, 4: {}, 5: {}}

        Attributes and custom IDs may be specified using a 3-tuple for each simplex.

        >>> S = xgi.SimplicialComplex()
        >>> simplices = [
        ...     ([0, 1], 'one', {'color': 'red'}),
        ...     ([1, 2], 'two', {'age': 30}),
        ...     ([2, 3, 4], 'three', {'color': 'blue', 'age': 40}),
        ... ]
        >>> S.add_simplices_from(simplices)
        >>> {e: S.edges[e] for e in S.edges}
        {'one': {'color': 'red'}, 'two': {'age': 30}, 'three': {'color': 'blue', 'age': 40}, 0: {}, 1: {}, 2: {}}

        """

        # format 5 is the easiest one
        if isinstance(ebunch_to_add, dict):

            faces = []  # container to store subfaces
            for id, members in ebunch_to_add.items():

                # check that it does not exist yet (based on members, not ID)
                if not members or self.has_simplex(members):
                    continue

                if id in self._edge.keys():  # check that uid is not present yet
                    warn(f"uid {id} already exists, cannot add simplex {members}.")
                    continue

                if max_order != None:
                    if len(members) > max_order + 1:
                        combos = powerset(members, include_singletons=False)
                        faces += list(combos)

                        continue

                try:
                    _ = frozenset(members)
                except TypeError as e:
                    raise XGIError("Invalid ebunch format") from e

                self._add_simplex(frozenset(members), id)

                update_uid_counter(self, id)

                # store subfaces
                faces += self._subfaces(members)

            # add subfaces
            faces = set(faces)  # get unique subfaces
            for members in faces:
                # check that it does not exist yet (based on members, not ID)
                if not members or self.has_simplex(members):
                    continue

                self._add_face(members)

            return

        # in formats 1-4 we only know that ebunch_to_add is an iterable, so we iterate
        # over it and use the firs element to determine which format we are working with
        new_edges = iter(ebunch_to_add)
        try:
            first_edge = next(new_edges)
        except StopIteration:
            return
        try:
            first_elem = list(first_edge)[0]
        except TypeError:
            first_elem = None

        format1, format2, format3, format4 = False, False, False, False
        if isinstance(first_elem, Iterable):
            if all(isinstance(e, str) for e in first_edge):
                format1 = True
            elif len(first_edge) == 2 and issubclass(type(first_edge[1]), Hashable):
                format2 = True
            elif len(first_edge) == 2:
                format3 = True
            elif len(first_edge) == 3:
                format4 = True
        else:
            format1 = True

        if (format1 and isinstance(first_edge, str)) or (
            not format1 and isinstance(first_elem, str)
        ):
            raise XGIError("Members cannot be specified as a string")

        faces = []
        # now we may iterate over the rest
        e = first_edge
        while True:
            if format1:
                members, id, eattr = e, None, {}  # uid now set below
            elif format2:
                members, id, eattr = e[0], e[1], {}
            elif format3:
                members, id, eattr = e[0], None, e[1]  # uid now set below
            elif format4:
                members, id, eattr = e[0], e[1], e[2]

            # check if members is iterable before checking it exists
            # to raise meaningful error if not iterable
            try:
                _ = iter(members)
            except TypeError as e:
                raise XGIError("Invalid ebunch format") from e

            # check that it does not exist yet (based on members, not ID)
            if not members or self.has_simplex(members):
                try:
                    e = next(new_edges)
                except StopIteration:
                    break

                continue

            # needs to go after the check for existence, otherwise
            # we're skipping ID numbers when edges already exist
            if format1 or format3:
                id = next(self._edge_uid)

            if max_order != None:
                if len(members) > max_order + 1:
                    combos = powerset(members, include_singletons=False)
                    faces += list(combos)  # store faces

                    try:
                        e = next(new_edges)
                    except StopIteration:
                        break

                    continue

            if id in self._edge.keys():  # check that uid is not present yet
                warn(f"uid {id} already exists, cannot add simplex {members}.")

                try:
                    e = next(new_edges)
                except StopIteration:
                    break

                continue

            try:
                self._edge[id] = frozenset(members)
            except TypeError as e:
                raise XGIError("Invalid ebunch format") from e

            for n in members:
                if n not in self._node:
                    self._node[n] = set()
                    self._node_attr[n] = self._node_attr_dict_factory()
                self._node[n].add(id)

            self._edge_attr[id] = self._hyperedge_attr_dict_factory()
            self._edge_attr[id].update(attr)
            self._edge_attr[id].update(eattr)

            update_uid_counter(self, id)

            # store subfaces
            faces += self._subfaces(members)

            try:
                e = next(new_edges)
            except StopIteration:
                break

        # add subfaces
        faces = set(faces)  # get unique faces
        for members in faces:

            # check that it does not exist yet (based on members, not ID)
            if not members or self.has_simplex(members):
                continue

            self._add_face(members)

    def close(self):
        """Adds all missing subfaces to the complex.

        See Also
        --------
        add_simplex : add a single simplex
        add_weighted_simplices_from : convenient way to add weighted simplices

        Notes
        -----
        Adding the same simplex twice will add it only once. Currently
        cannot add empty simplices; the method skips over them.
        """
        ebunch_to_close = list(map(list, self.edges.members()))
        for simplex in ebunch_to_close:
            if isinstance(simplex[-1], dict):
                dd = simplex[-1]
                simplex = simplex[:-1]
            else:
                dd = {}

            if simplex:
                new_faces = self._subfaces(simplex)
                self.add_simplices_from(new_faces)

    def add_weighted_simplices_from(
        self, ebunch_to_add, max_order=None, weight="weight", **attr
    ):
        """Add weighted simplices in `ebunch_to_add` with specified weight attr

        Parameters
        ----------
        ebunch_to_add : iterable of simplices
            Each simplex given in the list or container will be added
            to the graph. The simplices must be given as tuples of
            the form (node1, node2, ..., noden, weight).
        weight : string, optional (default= 'weight')
            The attribute name for the simplex weights to be added.
        attr : keyword arguments, optional (default= no attributes)
            simplex attributes to add/update for all simplices.

        See Also
        --------
        add_simplex : add a single simplex
        add_simplices_from : add multiple simplices

        Notes
        -----
        Adding the same simplex twice will add it only once.

        Example
        -------
        >>> import xgi
        >>> S = xgi.SimplicialComplex()
        >>> simplices = [(0, 1, 0.3), (0, 2, 0.8)]
        >>> S.add_weighted_simplices_from(simplices)
        >>> S.edges[0]
        {'weight': 0.3}
        """

        try:
            self.add_simplices_from(
                ((edge[:-1], {weight: edge[-1]}) for edge in ebunch_to_add),
                max_order=max_order,
                **attr,
            )
        except KeyError:
            XGIError("Empty or invalid simplices specified.")

    def _remove_simplex_id(self, id):
        """Helper function to remove a simplex with a given id

        Parameters
        ----------
        id : Hashable
            edge ID to remove

        """

        for node in self.edges.members(id):
            self._node[node].remove(id)
        del self._edge[id]
        del self._edge_attr[id]

    def remove_simplex_id(self, id):
        """Remove a simplex with a given id.

        This also removes all simplices of which this simplex is face,
        to preserve the simplicial complex structure.

        Parameters
        ----------
        id : Hashable
            edge ID to remove

        Raises
        ------
        XGIError
            If no edge has that ID.

        See Also
        --------
        remove_edges_from : remove a collection of edges
        """
        try:

            # remove all simplices that contain simplex
            supfaces_ids = self._supfaces_id(self._edge[id])
            for sup_id in supfaces_ids:
                self._remove_simplex_id(sup_id)

            # remove simplex
            self._remove_simplex_id(id)

        except KeyError as e:
            raise XGIError(f"Simplex {id} is not in the Simplicialcomplex") from e

    def remove_simplex_ids_from(self, ebunch):
        """Remove all simplicies specified in ebunch.

        Parameters
        ----------
        ebunch: list or iterable of hashables
            Each edge id given in the list or iterable will be removed
            from the Simplicialcomplex.

        Raises
        ------
        xgi.exception.IDNotFound
            If an id in ebunch is not part of the network.

        See Also
        --------
        remove_simplex_id : remove a single simplex by ID.

        """
        for id in ebunch:
            self.remove_simplex_id(id)

    def has_simplex(self, simplex):
        """Whether a simplex appears in the simplicial complex.

        Parameters
        ----------
        simplex : list or set
            An iterable of hashables that specifies an simplex

        Returns
        -------
        bool
           Whether or not simplex is as a simplex in the simplicial complex.

        Examples
        --------
        >>> import xgi
        >>> H = xgi.SimplicialComplex([[1, 2], [2, 3, 4]])
        >>> H.has_simplex([1, 2])
        True
        >>> H.has_simplex({1, 3})
        False

        """
        return frozenset(simplex) in self._edge.values()



"""Simulation of the Kuramoto model."""

import numpy as np

import xgi

from ..exception import XGIError

__all__ = [
    "simulate_kuramoto",
    "compute_kuramoto_order_parameter",
    "simulate_simplicial_kuramoto",
    "compute_simplicial_order_parameter",
]


def simulate_kuramoto(H, k2, k3, omega=None, theta=None, timesteps=10000, dt=0.002):
    """Simulates the Kuramoto model on hypergraphs.
    This solves the Kuramoto model ODE on hypergraphs with edges of sizes 2 and 3
    using the Euler Method. It returns timeseries of the phases.

    Parameters
    ----------
    H : Hypergraph object
        The hypergraph on which you run the Kuramoto model
    k2 : float
        The coupling strength for links
    k3 : float
        The coupling strength for triangles
    omega : numpy array of real values
        The natural frequency of the nodes. If None (default), randomly drawn from a normal distribution
    theta : numpy array of real values
        The initial phase distribution of nodes. If None (default), drawn from a random uniform distribution
        on [0, 2pi[.
    timesteps : int greater than 1, default: 10000
        The number of timesteps for Euler Method.
    dt : float greater than 0, default: 0.002
        The size of timesteps for Euler Method.

    Returns
    -------
    theta_time: numpy array of floats
        Timeseries of phases from the Kuramoto model, of dimension (N, T)
    times: numpy array of floats
        Times corresponding to the simulate phases

    References
    ----------
    "Synchronization of phase oscillators on complex hypergraphs"
    by Sabina Adhikari, Juan G. Restrepo and Per Sebastian Skardal
    https://doi.org/10.48550/arXiv.2208.00909

    Examples
    --------
    >>> import numpy as np
    >>> import xgi
    >>> n = 50
    >>> H = xgi.random_hypergraph(n, [0.05, 0.001], seed=None)
    >>> omega = 2*np.ones(n)
    >>> theta = np.linspace(0, 2*np.pi, n)
    >>> theta_time, times = simulate_kuramoto(H, k2=2, k3=3, omega=omega, theta=theta)

    """
    H_int = xgi.convert_labels_to_integers(H, "label")

    links = H_int.edges.filterby("size", 2).members()
    triangles = H_int.edges.filterby("size", 3).members()
    n = H_int.num_nodes

    theta_time = np.zeros((n, timesteps))
    times = np.arange(timesteps) * dt

    if omega is None:
        omega = np.random.normal(0, 1, n)

    if theta is None:
        theta = np.random.random(n) * 2 * np.pi

    for t in range(timesteps):

        theta_time[:, t] = theta

        r1 = np.zeros(n, dtype=complex)
        r2 = np.zeros(n, dtype=complex)

        for i, j in links:

            r1[i] += np.exp(1j * theta[j])
            r1[j] += np.exp(1j * theta[i])

        for i, j, k in triangles:

            r2[i] += np.exp(2j * theta[j] - 1j * theta[k]) + np.exp(
                2j * theta[k] - 1j * theta[j]
            )
            r2[j] += np.exp(2j * theta[i] - 1j * theta[k]) + np.exp(
                2j * theta[k] - 1j * theta[i]
            )
            r2[k] += np.exp(2j * theta[i] - 1j * theta[j]) + np.exp(
                2j * theta[j] - 1j * theta[i]
            )

        d_theta = (
            omega
            + k2 * np.multiply(r1, np.exp(-1j * theta)).imag
            + k3 * np.multiply(r2, np.exp(-1j * theta)).imag
        )
        theta_new = theta + d_theta * dt
        theta = theta_new

    return theta_time, times


def compute_kuramoto_order_parameter(theta_time):
    """This function calculates the order parameter for the Kuramoto model on hypergraphs,
    from time series, which is a measure of synchrony.

    Parameters
    ----------
    theta_time: numpy array of floats
        Timeseries of phases from the Kuramoto model, of dimension (N, T)

    Returns
    -------
    r_time : numpy array of floats
        Timeseries for Kuramoto model order parameter

    """

    z = np.mean(np.exp(1j * theta_time), axis=0)
    r_time = np.abs(z)

    return r_time


def simulate_simplicial_kuramoto(
    S,
    orientations=None,
    order=1,
    omega=[],
    sigma=1,
    theta0=[],
    T=10,
    n_steps=10000,
    index=False,
):
    """
    This function simulates the simplicial Kuramoto model's dynamics on an oriented simplicial complex
    using explicit Euler numerical integration scheme.

    Parameters
    ----------
    S: simplicial complex object
        The simplicial complex on which you
        run the simplicial Kuramoto model
    orientations: dict, Default : None
        Dictionary mapping non-singleton simplices IDs to their boolean orientation
    order: integer
        The order of the oscillating simplices
    omega: numpy.ndarray
        The simplicial oscillators' natural frequencies, has dimension
        (n_simplices of given order, 1)
    sigma: positive real value
        The coupling strength
    theta0: numpy.ndarray
        The initial phase distribution, has dimension
        (n_simplices of given order, 1)
    T: positive real value
        The final simulation time.
    n_steps: integer greater than 1
        The number of integration timesteps for
        the explicit Euler method.
    index: bool, default: False
        Specifies whether to output dictionaries mapping the node and edge IDs to indices

    Returns
    -------
    theta: numpy.ndarray
        Timeseries of the simplicial oscillators' phases, has dimension
        (n_simplices of given order, n_steps)
    theta_minus: numpy array of floats
        Timeseries of the projection of the phases onto lower order simplices,
        has dimension (n_simplices of given order - 1, n_steps)
    theta_plus: numpy array of floats
        Timeseries of the projection of the phases onto higher order simplices,
        has dimension (n_simplices of given order + 1, n_steps)
    om1_dict: dict
        The dictionary mapping indices to (order-1)-simplices IDs, if index is True
    o_dict: dict
        The dictionary mapping indices to (order)-simplices IDs, if index is True
    op1_dict: dict
        The dictionary mapping indices to (order+1)-simplices IDs, if index is True

    References
    ----------
    "Explosive Higher-Order Kuramoto Dynamics on Simplicial Complexes"
    by Ana P. Millán, Joaquín J. Torres, and Ginestra Bianconi
    https://doi.org/10.1103/PhysRevLett.124.218301

    """

    # Notation:
    # B_o - boundary matrix acting on (order)-simplices
    # D_o - adjoint boundary matrix acting on (order)-simplices
    # om1 = order - 1
    # op1 = order + 1

    if not isinstance(S, xgi.SimplicialComplex):
        raise XGIError(
            "The simplicial Kuramoto model can be simulated only on a SimplicialComplex object"
        )

    if index:
        B_o, om1_dict, o_dict = xgi.matrix.boundary_matrix(S, order, orientations, True)
    else:
        B_o = xgi.matrix.boundary_matrix(S, order, orientations, False)
    D_om1 = np.transpose(B_o)

    if index:
        B_op1, __, op1_dict = xgi.matrix.boundary_matrix(
            S, order + 1, orientations, True
        )
    else:
        B_op1 = xgi.matrix.boundary_matrix(S, order + 1, orientations, False)
    D_o = np.transpose(B_op1)

    # Compute the number of oscillating simplices
    n_o = np.shape(B_o)[1]

    dt = T / n_steps
    theta = np.zeros((n_o, n_steps))
    theta[:, [0]] = theta0
    for t in range(1, n_steps):
        theta[:, [t]] = theta[:, [t - 1]] + dt * (
            omega
            - sigma * D_om1 @ np.sin(B_o @ theta[:, [t - 1]])
            - sigma * B_op1 @ np.sin(D_o @ theta[:, [t - 1]])
        )
    theta_minus = B_o @ theta
    theta_plus = D_o @ theta
    if index:
        return theta, theta_minus, theta_plus, om1_dict, o_dict, op1_dict
    else:
        return theta, theta_minus, theta_plus


def compute_simplicial_order_parameter(theta_minus, theta_plus):
    """
    This function computes the simplicial order parameter of a
    simplicial Kuramoto dynamics simulation.

    Parameters
    ----------
    theta_minus: numpy.ndarray
        Timeseries of the projection of the phases onto
        lower order simplices, has dimension
        (n_simplices of given order - 1, n_steps)

    theta_plus: numpy.ndarray
        Timeseries of the projection of the phases onto
        higher order simplices, has dimension
        (n_simplices of given order + 1, n_steps)

    Returns
    -------
    R : numpy.ndarray
        Timeseries of the simplicial order parameter,
        has dimension (1, n_steps)

    References
    ----------
    "Connecting Hodge and Sakaguchi-Kuramoto through a mathematical
    framework for coupled oscillators on simplicial complexes"
    by Alexis Arnaudon, Robert L. Peach, Giovanni Petri, and Paul Expert
    https://doi.org/10.1038/s42005-022-00963-7

    """

    C = np.size(theta_minus, 0) + np.size(theta_plus, 0)
    R = (np.sum(np.cos(theta_minus), 0) + np.sum(np.cos(theta_plus), 0)) / C
    return R



"""Read from and write to incidece matrices."""

import numpy as np

from ..convert import from_incidence_matrix
from ..linalg import incidence_matrix

__all__ = [
    "read_incidence_matrix",
    "write_incidence_matrix",
]


def read_incidence_matrix(
    path, comments="#", delimiter=None, create_using=None, encoding="utf-8"
):
    """Read a file containing an incidence matrix and
    convert it to a Hypergraph object.

    Parameters
    ----------
    path: string
        The path of the file to read from
    comments: string, default: "#"
        The token that denotes comments in the file
    delimiter: char, default: space (" ")
        Specifies the delimiter between hyperedge members
    create_using : Hypergraph constructor, optional
        The hypergraph object to add the data to, by default None
    nodetype: type
        type that the node labels will be cast to
    encoding: string, default: "utf-8"
        Encoding of the file

    Returns
    -------
    A Hypergraph object
        The loaded hypergraph

    See Also
    --------
    write_incidence_matrix

    Examples
    --------
    >>> import xgi
    >>> # H = xgi.read_incidence_matrix("test.csv", delimiter=",")

    """
    return from_incidence_matrix(
        np.loadtxt(path, comments=comments, delimiter=delimiter, encoding=encoding),
        create_using=create_using,
    )


def write_incidence_matrix(H, path, delimiter=" ", encoding="utf-8"):
    """Write a Hypergraph object to a file as an incidence matrix.

    Parameters
    ----------
    H: Hypergraph object
        The hypergraph of interest
    path: string
        The path of the file to write to
    delimiter: char, default: space (" ")
        Specifies the delimiter between hyperedge members
    encoding: string, default: "utf-8"
        Encoding of the file

    See Also
    --------
    read_incidence_matrix

    Examples
    --------
    >>> import xgi
    >>> H = xgi.random_hypergraph(50, [0.01, 0.001])
    >>> # xgi.write_incidence_matrix(H, "test.csv", delimiter=",")

    """
    I = incidence_matrix(H, sparse=False)
    np.savetxt(path, I, delimiter=delimiter, newline="\n", encoding=encoding)



import json
import os
from functools import lru_cache
from warnings import warn

import requests

from .. import convert
from ..exception import XGIError

__all__ = ["load_xgi_data", "download_xgi_data"]


def load_xgi_data(
    dataset,
    cache=True,
    read=False,
    path="",
    nodetype=None,
    edgetype=None,
    max_order=None,
):
    """Load a data set from the xgi-data repository or a local file.

    Parameters
    ----------
    dataset : str
        Dataset name. Valid options are the top-level tags of the
        index.json file in the xgi-data repository.
    cache : bool, optional
        Whether to cache the input data
    read : bool, optional
        If read==True, search for a local copy of the data set. Use the local
        copy if it exists, otherwise use the  xgi-data repository.
    path : str, optional
        Path to a local copy of the data set
    nodetype : type, optional
        Type to cast the node ID to
    edgetype : type, optional
        Type to cast the edge ID to
    max_order: int, optional
        Maximum order of edges to add to the hypergraph

    Returns
    -------
    Hypergraph
        The loaded hypergraph.

    Raises
    ------
    XGIError
       The specified dataset does not exist.
    """

    if read:
        cfp = os.path.join(path, dataset + ".json")
        if os.path.exists(cfp):
            data = json.load(open(cfp, "r"))

            return convert.dict_to_hypergraph(
                data, nodetype=nodetype, edgetype=edgetype, max_order=max_order
            )
        else:
            warn(
                f"No local copy was found at {cfp}. The data is requested from the xgi-data repository instead. To download a local copy, use `download_xgi_data`."
            )
    if cache:
        data = _request_from_xgi_data_cached(dataset)
    else:
        data = _request_from_xgi_data(dataset)

    return convert.dict_to_hypergraph(
        data, nodetype=nodetype, edgetype=edgetype, max_order=max_order
    )


def download_xgi_data(dataset, path=""):
    """Make a local copy of a dataset in the xgi-data repository.

    Parameters
    ----------
    dataset : str
        Dataset name. Valid options are the top-level tags of the
        index.json file in the xgi-data repository.

    path : str, optional
        Path to where the local copy should be saved. If none is given, save
        file to local directory.
    """

    jsondata = _request_from_xgi_data(dataset)
    jsonfile = open(os.path.join(path, dataset + ".json"), "w")
    json.dump(jsondata, jsonfile)
    jsonfile.close()


def _request_from_xgi_data(dataset):
    """Request a dataset from xgi-data.

    Parameters
    ----------
    dataset : str
        Dataset name. Valid options are the top-level tags of the
        index.json file in the xgi-data repository.

    Returns
    -------
    Data
        The requested data loaded from a json file.

    Raises
    ------
    XGIError
        If HTTP request is not successful of the dataset does not exist.

    See also
    ---------
    load_xgi_data
    """

    index_url = "https://gitlab.com/complexgroupinteractions/xgi-data/-/raw/main/index.json?inline=false"
    r = requests.get(index_url)

    if r.ok:
        index_data = r.json()
    else:
        raise XGIError(f"Error: HTTP response {r.status_code}")

    key = dataset.lower()
    if key not in index_data:
        print("Valid dataset names:")
        print(*index_data, sep="\n")
        raise XGIError("Must choose a valid dataset name!")

    r = requests.get(index_data[key]["url"])
    if r.ok:
        return r.json()
    else:
        raise XGIError(f"Error: HTTP response {r.status_code}")


@lru_cache(maxsize=None)
def _request_from_xgi_data_cached(dataset):
    """Request a dataset from xgi-data and cache the result.

    Wraps `_request_from_xgi_data` in an lru_cache decorator.

    Parameters
    ----------
    dataset : str
        Dataset name. Valid options are the top-level tags of the
        index.json file in the xgi-data repository.

    Returns
    -------
    Data
        The requested data loaded from a json file.

    See also
    ---------
    load_xgi_data
    """

    return _request_from_xgi_data(dataset)



"""Read from and write to bipartite formats."""

from ..exception import XGIError
from ..generators import empty_hypergraph

__all__ = [
    "read_bipartite_edgelist",
    "write_bipartite_edgelist",
    "parse_bipartite_edgelist",
]


def generate_bipartite_edgelist(H, delimiter=" "):
    """
    A helper function to generate a bipartite edge list from a Hypergraph object.

    Parameters
    ----------
    H: Hypergraph object
        The hypergraph of interest
    delimiter: char, default: space (" ")
        Specifies the delimiter between hyperedge members

    Yields
    -------
    A iterator of strings
        Each entry is a line to be written to the output file.
    """
    for id in H.edges:
        for node in H.edges.members(id):
            yield delimiter.join(map(str, [node, id]))


def write_bipartite_edgelist(H, path, delimiter=" ", encoding="utf-8"):
    """Write a Hypergraph object to a file
    as a bipartite edgelist.

    Parameters
    ----------
    H: Hypergraph object
        The hypergraph of interest
    path: string
        The path of the file to write to
    delimiter: char, default: space (" ")
        Specifies the delimiter between hyperedge members
    encoding: string, default: "utf-8"
        Encoding of the file

    See Also
    --------
    read_bipartite_edgelist

    Example
    -------
    >>> import xgi
    >>> H = xgi.random_hypergraph(50, [0.01, 0.001])
    >>> # xgi.write_bipartite_edgelist(H, "test.csv", delimiter=",")

    """
    with open(path, "wb") as file:
        for line in generate_bipartite_edgelist(H, delimiter):
            line += "\n"
            file.write(line.encode(encoding))


def read_bipartite_edgelist(
    path,
    comments="#",
    delimiter=None,
    create_using=None,
    nodetype=None,
    edgetype=None,
    dual=False,
    encoding="utf-8",
):
    """Read a file containing a bipartite edge list and
    convert it to a Hypergraph object.

    Parameters
    ----------
    path: string
        The path of the file to read from
    comments: string, default: "#"
        The token that denotes comments in the file
    delimiter: char, default: space (" ")
        Specifies the delimiter between hyperedge members
    create_using : Hypergraph constructor, optional
        The hypergraph object to add the data to, by default None
    nodetype: type
        type that the node labels will be cast to
    edgetype: type
        type that the edge labels will be cast to
    dual: bool, default: False
        Specifies whether the node IDs are in the second column. If False,
        the node IDs are in the first column.
    encoding: string, default: "utf-8"
        Encoding of the file

    Returns
    -------
    A Hypergraph object
        The loaded hypergraph

    See Also
    --------
    write_bipartite_edgelist

    Example
    -------
    >>> import xgi
    >>> # H = xgi.read_bipartite_edgelist("test.csv", delimiter=",")

    """
    with open(path, "rb") as file:
        lines = (
            line if isinstance(line, str) else line.decode(encoding) for line in file
        )
        return parse_bipartite_edgelist(
            lines,
            comments=comments,
            delimiter=delimiter,
            create_using=create_using,
            nodetype=nodetype,
            edgetype=edgetype,
            dual=dual,
        )


def parse_bipartite_edgelist(
    lines,
    comments="#",
    delimiter=None,
    create_using=None,
    nodetype=None,
    edgetype=None,
    dual=False,
):
    """
    A helper function to read a iterable of strings containing a bipartite edge list and
    convert it to a Hypergraph object.

    Reads the first two entries of each line and assumes that the first entry is a node
    ID and that the second entry is an edge ID. Raises error if there are fewer than two
    entries.

    Parameters
    ----------
    lines: iterable of strings
        Lines where each line is a bipartite edge
    comments: string, default: "#"
        The token that denotes comments to ignore
    delimiter: char, default: space (" ")
        Specifies the delimiter between hyperedge members
    create_using : Hypergraph constructor, optional
        The hypergraph object to add the data to, by default None
    nodetype: type
        type that the node labels will be cast to
    edgetype: type
        type that the edge labels will be cast to
    data: bool, default: False
        Specifies whether there is a dictionary of data at the end of the line.

    Raises
    ------
    XGIError
        If a line contains fewer than two entries
    TypeError
        If node types fail to be converted

    Returns
    -------
    Hypergraph
        The loaded hypergraph.

    """
    H = empty_hypergraph(create_using)

    node_index = 1 if dual else 0
    edge_index = 0 if dual else 1

    for line in lines:
        if comments is not None:
            p = line.find(comments)
            if p >= 0:
                line = line[:p]
            if not line:
                continue
        s = line.strip().split(delimiter)
        if len(s) < 2:
            raise XGIError("Each line must contain at least two entries!")
        # no data or data type specified

        # convert node types
        if nodetype is not None:
            try:
                node = nodetype(s[node_index])
            except ValueError as e:
                raise TypeError(
                    f"Failed to convert the node with ID {s[node_index]} to type {nodetype}."
                ) from e
        else:
            node = s[node_index]

        # convert edge types
        if edgetype is not None:
            try:
                edge = edgetype(s[edge_index])
            except ValueError as e:
                raise TypeError(
                    f"Failed to convert the edge with ID {s[edge_index]} to type {edgetype}."
                ) from e
        else:
            edge = s[edge_index]

        H.add_node_to_edge(edge, node)
    return H



from collections import defaultdict
from copy import deepcopy

import networkx as nx
import pandas as pd
from networkx.algorithms import bipartite
from numpy import matrix, ndarray
from scipy.sparse import coo_matrix, csc_matrix, csr_matrix, lil_matrix

from .classes import (
    Hypergraph,
    SimplicialComplex,
    maximal_simplices,
    set_edge_attributes,
)
from .exception import XGIError
from .generators import empty_hypergraph, empty_simplicial_complex
from .linalg import adjacency_matrix, incidence_matrix
from .utils.utilities import dual_dict

__all__ = [
    "convert_to_hypergraph",
    "convert_to_graph",
    "convert_to_simplicial_complex",
    "from_hyperedge_list",
    "to_hyperedge_list",
    "from_hyperedge_dict",
    "to_hyperedge_dict",
    "from_bipartite_pandas_dataframe",
    "to_bipartite_pandas_dataframe",
    "from_incidence_matrix",
    "from_simplicial_complex_to_hypergraph",
    "to_incidence_matrix",
    "from_bipartite_graph",
    "to_bipartite_graph",
    "dict_to_hypergraph",
]


def convert_to_hypergraph(data, create_using=None):
    """Make a hypergraph from a known data structure.

    The preferred way to call this is automatically from the class constructor.

    Parameters
    ----------
    data : object to be converted
        Current known types are:
         * a Hypergraph object
         * list-of-iterables
         * dict-of-iterables
         * Pandas DataFrame (bipartite edgelist)
         * numpy matrix
         * numpy ndarray
         * scipy sparse matrix
    create_using : Hypergraph constructor, optional (default=Hypergraph)
        Hypergraph type to create. If hypergraph instance, then cleared before populated.

    Returns
    -------
    Hypergraph object
        A hypergraph constructed from the data

    """
    if data is None:
        return empty_hypergraph(create_using)

    elif isinstance(data, Hypergraph):
        H = empty_hypergraph(create_using)
        H.add_nodes_from((n, attr) for n, attr in data.nodes.items())
        ee = data.edges
        H.add_edges_from((ee.members(e), e, deepcopy(attr)) for e, attr in ee.items())
        H._hypergraph = deepcopy(data._hypergraph)

    elif isinstance(data, list):
        # edge list
        from_hyperedge_list(data, create_using)

    elif isinstance(data, pd.DataFrame):
        from_bipartite_pandas_dataframe(data, create_using)

    elif isinstance(data, dict):
        # edge dict in the form we need
        from_hyperedge_dict(data, create_using)

    elif isinstance(
        data, (ndarray, matrix, csr_matrix, csc_matrix, coo_matrix, lil_matrix)
    ):
        from_incidence_matrix(data, create_using)

    else:
        raise XGIError("Input data has unsupported type.")


def convert_to_graph(H):
    """Graph projection (1-skeleton) of the hypergraph H.
    Weights are not considered.

    Parameters
    ----------
    H : Hypergraph object
        The hypergraph of interest

    Returns
    -------
    G : networkx.Graph
        The graph projection
    """

    A = adjacency_matrix(H)  # This is unweighted by design
    G = nx.from_scipy_sparse_matrix(A)
    G = nx.relabel_nodes(G, {i: node for i, node in enumerate(H.nodes)})
    return G


def convert_to_simplicial_complex(data, create_using=None):
    """Make a hypergraph from a known data structure.
    The preferred way to call this is automatically
    from the class constructor.

    Parameters
    ----------
    data : object to be converted
        Current known types are:
         * a Hypergraph object
         * list-of-iterables
         * dict-of-iterables
         * Pandas DataFrame (bipartite edgelist)
         * numpy matrix
         * numpy ndarray
         * scipy sparse matrix
    create_using : Hypergraph graph constructor, optional (default=Hypergraph)
        Hypergraph type to create. If hypergraph instance, then cleared before populated.

    Returns
    -------
    Hypergraph object
        A hypergraph constructed from the data
    """

    if data is None:
        return empty_hypergraph(create_using)

    elif isinstance(data, SimplicialComplex):
        H = empty_simplicial_complex(create_using)
        H.add_nodes_from((n, attr) for n, attr in data.nodes.items())
        ee = data.edges
        H.add_simplices_from(
            (ee.members(e), e, deepcopy(attr)) for e, attr in ee.items()
        )
        H._hypergraph = deepcopy(data._hypergraph)

    elif isinstance(data, list):
        # edge list
        from_hyperedge_list(data, create_using)

    elif isinstance(data, pd.DataFrame):
        from_bipartite_pandas_dataframe(data, create_using)

    elif isinstance(data, dict):
        # edge dict in the form we need
        raise XGIError("Cannot generate SimplicialComplex from simplex dictionary")
    elif isinstance(
        data, (ndarray, matrix, csr_matrix, csc_matrix, coo_matrix, lil_matrix)
    ):
        # incidence matrix
        raise XGIError(
            "Not implemented: construction of a SimplicialComplex from incidence matrix"
        )
    else:
        raise XGIError("Input data has unsupported type.")


def from_hyperedge_list(d, create_using=None, max_order=None):
    """Generate a hypergraph from a list of lists.

    Parameters
    ----------
    d : list of iterables
        A hyperedge list
    create_using : Hypergraph constructor, optional
        The hypergraph to add the edges to, by default None

    Returns
    -------
    Hypergraph object
        The constructed hypergraph object

    See Also
    --------
    to_hyperedge_list
    """
    H = empty_hypergraph(create_using)
    if isinstance(H, SimplicialComplex):
        H.add_simplices_from(d, max_order=max_order)
    else:
        H.add_edges_from(d)
    return H


def to_hyperedge_list(H):
    """Generate a hyperedge list from a hypergraph.

    Parameters
    ----------
    H : Hypergraph object
        The hypergraph of interest

    Returns
    -------
    list of sets
        The hyperedge list

    See Also
    --------
    from_hyperedge_list
    """
    return H.edges.members()


def from_hyperedge_dict(d, create_using=None):
    """Creates a hypergraph from a dictionary of hyperedges

    Parameters
    ----------
    d : dict
        A dictionary where the keys are edge IDs and the values
        are containers of nodes specifying the edges.
    create_using : Hypergraph constructor, optional
        The hypergraph object to add the data to, by default None

    Returns
    -------
    Hypergraph object
        The constructed hypergraph object

    See Also
    --------
    to_hyperedge_dict

    """
    H = empty_hypergraph(create_using)
    H.add_nodes_from(dual_dict(d))
    H.add_edges_from((members, uid) for uid, members in d.items())
    return H


def to_hyperedge_dict(H):
    """Outputs a hyperedge dictionary

    Parameters
    ----------
    H : Hypergraph object
        The hypergraph of interest

    Returns
    -------
    dict
        A dictionary where the keys are edge IDs and the values
        are sets of nodes specifying the edges.

    See Also
    --------
    from_hyperedge_dict
    """
    return deepcopy(H._edge)


def from_bipartite_pandas_dataframe(
    df, create_using=None, node_column=0, edge_column=1
):
    """Create a hypergraph from a pandas dataframe given
    specified node and edge columns.

    Parameters
    ----------
    df : Pandas dataframe
        A dataframe where specified columns list the node IDs
        and the associated edge IDs
    create_using : Hypergraph constructor, optional
        The hypergraph object to add the data to, by default None
    node_column : hashable, optional
        The column with the node IDs, by default 0
        Can specify names or indices
    edge_column : hashable, optional
        The column with the edge IDs, by default 1
        Can specify names or indices

    Returns
    -------
    Hypergraph object
        The constructed hypergraph

    Raises
    ------
    XGIError
        Raises an error if the user specifies invalid column names
    """
    H = empty_hypergraph(create_using)

    # try to get by labels first
    try:
        d = df[[node_column, edge_column]]
    except KeyError:
        # try to index the labels
        try:
            columns = list(df.columns)
            d = df[[columns[node_column], columns[edge_column]]]
        except (KeyError, TypeError):
            raise XGIError("Invalid columns specified")

    if isinstance(H, SimplicialComplex):
        simplex_list = defaultdict(list)
        for line in d.itertuples(index=False):
            if line[0] not in simplex_list[line[1]]:
                simplex_list[line[1]].append(line[0])

        H.add_simplices_from(list(simplex_list.values()))
    else:
        for line in d.itertuples(index=False):
            node = line[0]
            edge = line[1]
            H.add_node_to_edge(edge, node)

    return H


def to_bipartite_pandas_dataframe(H):
    """Create a two column dataframe from a hypergraph.

    Parameters
    ----------
    H : Hypergraph or Simplicial Complex
        A dataframe where specified columns list the node IDs
        and the associated edge IDs

    Returns
    -------
    Pandas Dataframe object
        A two column dataframe

    Raises
    ------
    XGIError
        Raises an error if the user specifies invalid column names
    """
    data = []
    for id1, members in H._node.items():
        for id2 in members:
            data.append([id1, id2])
    return pd.DataFrame(data, columns=["Node ID", "Edge ID"])


def from_incidence_matrix(d, create_using=None, nodelabels=None, edgelabels=None):
    """Create a hypergraph from an incidence matrix

    Parameters
    ----------
    d : numpy array or a scipy sparse arrary
        The incidence matrix where rows specify nodes and columns specify edges.
    create_using : Hypergraph constructor, optional
        The hypergraph object to add the data to, by default None
    nodelabels : list or 1D numpy array, optional
        The ordered list of node IDs to map the indices
        of the incidence matrix to, by default None
    edgelabels : list or 1D numpy array, optional
        The ordered list of edge IDs to map the indices
        of the incidence matrix to, by default None

    Returns
    -------
    Hypergraph object
        The constructed hypergraph

    Raises
    ------
    XGIError
        Raises an error if the specified labels are the wrong dimensions

    See Also
    --------
    incidence_matrix
    to_incidence_matrix
    """
    I = coo_matrix(d)
    n, m = I.shape

    if nodelabels is None:
        nodedict = dict(zip(range(n), range(n)))
    elif nodelabels is not None and len(nodelabels) != n:
        raise XGIError("Node dictionary is the wrong size.")
    else:
        nodedict = dict(zip(range(n), nodelabels))

    if edgelabels is None:
        edgedict = dict(zip(range(m), range(m)))
    elif edgelabels is not None and len(edgelabels) != m:
        raise XGIError("Edge dictionary is the wrong size.")
    else:
        edgedict = dict(zip(range(m), edgelabels))

    H = empty_hypergraph(create_using)

    for node, edge in zip(I.row, I.col):
        node = nodedict[node]
        edge = edgedict[edge]
        H.add_node_to_edge(edge, node)

    return H


def from_simplicial_complex_to_hypergraph(SC):
    """Returns a hypergraph constructed from the
    maximal simpices of the provided simplicial complex.

    Parameters
    ----------
    SC : SimplicialComplex

    Returns
    -------
    Hypergraph

    """
    if type(SC) != SimplicialComplex:
        raise XGIError("The input must be a SimplicialComplex")

    max_simplices = maximal_simplices(SC)
    H = Hypergraph()
    H.add_edges_from([list(SC.edges.members(e)) for e in max_simplices])
    return H


def to_incidence_matrix(H, sparse=True, index=False):
    """Convert a hypergraph to an incidence matrix.

    Parameters
    ----------
    H : Hypergraph object
        The hypergraph of interest
    sparse : bool, optional
        Whether the constructed incidence matrix
        should be sparse, by default True
    index : bool, optional
        Whether to return the corresponding
        node and edge labels, by default False

    Returns
    -------
    numpy.ndarray or scipy csr_matrix
        The incidence matrix
    dict
        The dictionary mapping indices to node IDs, if index is True
    dict
        The dictionary mapping indices to edge IDs, if index is True

    See Also
    --------
    incidence_matrix
    from_incidence_matrix
    """
    return incidence_matrix(H, sparse=sparse, index=index)


def from_bipartite_graph(G, create_using=None, dual=False):
    """
    Create a Hypergraph from a NetworkX bipartite graph.

    Any hypergraph may be represented as a bipartite graph where
    nodes in the first layer are nodes and nodes in the second layer
    are hyperedges.

    The default behavior is to create nodes in the hypergraph
    from the nodes in the bipartite graph where the attribute
    bipartite=0 and hyperedges in the hypergraph from the nodes
    in the bipartite graph with attribute bipartite=1. Setting the
    keyword `dual` reverses this behavior.


    Parameters
    ----------
    G : nx.Graph
        A networkx bipartite graph. Each node in the graph has a property
        'bipartite' taking the value of 0 or 1 indicating the type of node.

    create_using : Hypergraph constructor, optional
        The hypergraph object to add the data to, by default None

    dual : bool, default : False
        If True, get edges from bipartite=0 and nodes from bipartite=1

    Returns
    -------
    Hypergraph
        The equivalent hypergraph

    References
    ----------
    The Why, How, and When of Representations for Complex Systems,
    Leo Torres, Ann S. Blevins, Danielle Bassett, and Tina Eliassi-Rad,
    https://doi.org/10.1137/20M1355896

    Examples
    --------
    >>> import networkx as nx
    >>> import xgi
    >>> G = nx.Graph()
    >>> G.add_nodes_from([1, 2, 3, 4], bipartite=0)
    >>> G.add_nodes_from(['a', 'b', 'c'], bipartite=1)
    >>> G.add_edges_from([(1, 'a'), (1, 'b'), (2, 'b'), (2, 'c'), (3, 'c'), (4, 'a')])
    >>> H = xgi.from_bipartite_graph(G)

    """
    edges = []
    nodes = []
    for n, d in G.nodes(data=True):
        try:
            node_type = d["bipartite"]
        except KeyError as e:
            raise XGIError("bipartite property not set") from e

        if node_type == 0:
            nodes.append(n)
        elif node_type == 1:
            edges.append(n)
        else:
            raise XGIError("Invalid type specifier")

    if not bipartite.is_bipartite_node_set(G, nodes):
        raise XGIError("The network is not bipartite")

    H = empty_hypergraph(create_using)
    H.add_nodes_from(nodes)
    for edge in edges:
        nodes_in_edge = list(G.neighbors(edge))
        H.add_edge(nodes_in_edge, id=edge)
    return H.dual() if dual else H


def to_bipartite_graph(H):
    """
    Create a NetworkX bipartite network from a hypergraph.

    Parameters
    ----------
    H: xgi.Hypergraph
        The XGI hypergraph object of interest

    Returns
    -------
    nx.Graph
        The resulting equivalent bipartite graph

    References
    ----------
    The Why, How, and When of Representations for Complex Systems,
    Leo Torres, Ann S. Blevins, Danielle Bassett, and Tina Eliassi-Rad,
    https://doi.org/10.1137/20M1355896

    Examples
    --------
    >>> import xgi
    >>> hyperedge_list = [[1, 2], [2, 3, 4]]
    >>> H = xgi.Hypergraph(hyperedge_list)
    >>> G = xgi.to_bipartite_graph(H)
    """
    G = nx.Graph()

    node_dict = dict(zip(H.nodes, range(H.num_nodes)))
    edge_dict = dict(zip(H.edges, range(H.num_nodes, H.num_nodes + H.num_edges)))
    G.add_nodes_from(node_dict.values(), bipartite=0)
    G.add_nodes_from(edge_dict.values(), bipartite=1)
    for node in H.nodes:
        for edge in H.nodes.memberships(node):
            G.add_edge(node_dict[node], edge_dict[edge])

    return (
        G,
        dict(zip(range(H.num_nodes), H.nodes)),
        dict(zip(range(H.num_nodes, H.num_nodes + H.num_edges), H.edges)),
    )


def dict_to_hypergraph(data, nodetype=None, edgetype=None, max_order=None):
    """
    A function to read a file in a standardized JSON format.

    Parameters
    ----------
    data: dict
        A dictionary in the hypergraph JSON format
    nodetype: type, optional
        Type that the node IDs will be cast to
    edgetype: type, optional
        Type that the edge IDs will be cast to
    max_order: int, optional
        Maximum order of edges to add to the hypergraph

    Returns
    -------
    A Hypergraph object
        The loaded hypergraph

    Raises
    ------
    XGIError
        If the JSON is not in a format that can be loaded.

    See Also
    --------
    read_json

    """
    H = empty_hypergraph()
    try:
        H._hypergraph.update(data["hypergraph-data"])
    except KeyError:
        raise XGIError("Failed to get hypergraph data attributes.")

    try:
        for id, dd in data["node-data"].items():
            if nodetype is not None:
                try:
                    id = nodetype(id)
                except ValueError as e:
                    raise TypeError(
                        f"Failed to convert edge IDs to type {nodetype}."
                    ) from e
            H.add_node(id, **dd)
    except KeyError:
        raise XGIError("Failed to import node attributes.")

    try:
        for id, edge in data["edge-dict"].items():
            if max_order and len(edge) > max_order + 1:
                continue
            if edgetype is not None:
                try:
                    id = edgetype(id)
                except ValueError as e:
                    raise TypeError(
                        f"Failed to convert the edge with ID {id} to type {edgetype}."
                    ) from e

            if nodetype is not None:
                try:
                    edge = {nodetype(n) for n in edge}
                except ValueError as e:
                    raise TypeError(
                        f"Failed to convert nodes to type {nodetype}."
                    ) from e
            H.add_edge(edge, id)

    except KeyError as e:
        raise XGIError("Failed to import edge dictionary.") from e

    try:
        if edgetype is None:
            edge_data = {
                key: val for key, val in data["edge-data"].items() if key in H.edges
            }
        else:
            edge_data = {
                edgetype(e): dd
                for e, dd in data["edge-data"].items()
                if edgetype(e) in H.edges
            }

        set_edge_attributes(
            H,
            edge_data,
        )
    except KeyError as e:
        raise XGIError("Failed to import edge attributes.") from e

    return H



class XGIException(Exception):
    """Base class for exceptions in XGI."""


class XGIError(XGIException):
    """Exception for a serious error in XGI"""


class IDNotFound(KeyError):
    """Raised when a node or edge is not in the hypergraph."""
